# -*- coding: utf-8 -*-

#############################################################################
# Copyright (c) 2014-2017 Redfern Electronics Limited. All rights reserved. #
#############################################################################

import canvas
import wx
import string
try:
    from pubsub import pub
except ImportError:
    from wx.lib.pubsub import pub
from lang import _

import xml.etree.cElementTree as ET

event_colour = (0xFF,0x5E,0x3A,0xFF)
io_colour = (0x00,0x7A,0xFF,0xFF)
crumb_colour = (0xB6,0xAE,0xA3,0xFF)
sparkle_colour = (0x58,0x56,0xD6,0xFF)
control_colour = (0xFF,0x95,0x00,0xFF)
data_colour = (0xC8,0x6E,0xDF,0xFF)
operator_colour = (0x4C,0xD9,0x64,255)
music_colour = (0x10,0x10,0x10,0xFF)

def from_xml(element, canvas_dest):
    import statements
    import hardware
    name = element.tag
    modules = [statements, hardware]
    module = None
    for test_module in modules:
        if hasattr(test_module, name):
            module = test_module

    if module:
        x_element = element.find('x')
        y_element = element.find('y')
        if x_element is not None:
            x = int(x_element.text)
        else:
            x = 0
        if y_element is not None:
            y = int(y_element.text)
        else:
            y = 0

        new_block = getattr(module, name)(x, y)
        canvas_dest.add(new_block)
        new_block.from_xml(element, canvas_dest)
        return new_block
    else:
        raise ValueError


class XMLBlock(object):
    def to_xml(self, name=None, position=False):
        if name==None:
            element = ET.Element(self.__class__.__name__)
        else:
            element = ET.Element(name)
        if position:
            ET.SubElement(element, 'x').text = str(int(self.x))
            ET.SubElement(element, 'y').text = str(int(self.y))

        for i, attach_point in enumerate(self.attach_points):
            if attach_point.attached_to:
                attach_point_element = ET.SubElement(element, 'attach_point', {'index': str(i)})
                attach_point_element.append(attach_point.attached_to.to_xml())
        for i, adopt_point in enumerate(self.adopt_points):
            if adopt_point.attached_to:
                adopt_point_element = ET.SubElement(element, 'adopt_point', {'index': str(i)})
                adopt_point_element.append(adopt_point.attached_to.to_xml())

        return element

    def from_xml(self, element, canvas_dest):
        for attach_point_element in element.findall('attach_point'):
            attached_block = from_xml(attach_point_element[0], canvas_dest)
            if attached_block:
                canvas_dest.float_all_attached(attached_block)
                canvas_dest.attach_blocks_to_point(self.attach_points[0])

        for adopt_point_element in element.findall('adopt_point'):
            index = int(adopt_point_element.attrib['index'])
            attached_block = from_xml(adopt_point_element[0], canvas_dest)
            if attached_block:
                canvas_dest.float_all_attached(attached_block)
                canvas_dest.adopt_floating_to_point(self.adopt_points[index])


class Button(canvas.Block):
    def __init__(self, x, y, label, colour):
        super(Button, self).__init__(x, y)
        self.colour = colour
        self._highlighted = False
        self.text = canvas.Text(0, 0, label, colour)
        self.outline = canvas.Rectangle(0, 0, 0, 0, 0, fill=(255,255,255,255), stroke=colour)
        self.add(self.outline)
        self.add(self.text)
        self.label = label

    @property
    def highlighted(self):
        return self._highlighted

    @highlighted.setter
    def highlighted(self, state):
        self._highlighted = state
        if state:
            self.outline.fill = self.colour
            self.text.fill = (0xFF,0xFF,0XFF,0xFF)
        else:
            self.outline.fill = (0xFF,0xFF,0XFF,0xFF)
            self.text.fill = self.colour
        self.render()

    @property
    def label(self):
        return self._label

    @label.setter
    def label(self, new_label):
        self._label = new_label
        self.text.text = new_label
        self.render()

    def layout(self):
        box = self.text.box
        radius = box.height / 2 + 2
        self.text.x = radius
        self.text.y = 2
        self.outline.radius = radius
        self.outline.width = box.width + 2 * radius
        self.outline.height = 2 * radius


class ToggledEntryBlock(canvas.Block, XMLBlock):
    "Data entry block that cycled through options on click."
    def __init__(self, x, y, options=['HI', 'LO']):
        super(ToggledEntryBlock, self).__init__(x, y)
        self.index = 0
        self.options = options
        self.text = canvas.Text(0, 0, _(options[self.index]))
        self.outline = canvas.Rectangle(0, 0, 0, 0, 0, fill=(0xe0,0xe0,0xe0,0xff), stroke=(0xe0,0xe0,0xe0,0xff))
        self.add(self.outline)
        self.add(self.text)
        self.render()

    @property
    def option(self):
        return self.options[self.index]

    @option.setter
    def option(self, option):
        if option in self.options:
            self.index = self.options.index(option)
            self.text.text = _(self.options[self.index])
            self.render()

    def layout(self):
        box = self.text.box
        radius = box.height / 2
        self.text.x = radius
        self.text.y = 0
        self.outline.radius = radius
        self.outline.width = box.width + 2 * radius
        self.outline.height = box.height

    def on_click(self):
        self.index += 1
        if self.index == len(self.options):
            self.index = 0
        self.text.text = _(self.options[self.index])
        self.render()

    def to_xml(self, name=None, position=False):
        element = XMLBlock.to_xml(self, name, position)
        element.text = self.option
        return element

    def from_xml(self, element, canvas_dest):
        if element is not None:
            XMLBlock.from_xml(self, element, canvas_dest)
            self.option = element.text


class BooleanEntryBlock(canvas.Block, XMLBlock):
    def __init__(self, x, y):
        super(BooleanEntryBlock, self).__init__(x, y)
        self.outline = canvas.RightAngledHexagon(0, 0, 32, 16, fill=(0xFF,0xFF,0xFF,0xFF), stroke=(0xFF,0xFF,0xFF,0xFF), stroke_width=1.0)
        self.add(self.outline)
        self.adopt_point = canvas.AttachPoint(0, 0, ['Bool'], self)
        self.adopt_points = [self.adopt_point]
        self.render()

    def layout(self):
        if self.adopt_point.attached_to:
            self.outline.visible = False
            width, height = self.distribute_hozontally(self.children[1:])
        else:
            self.outline.visible = True
            width, height = self.distribute_hozontally([self.outline])
        self.adopt_point.y = height/2

    def basic_code(self, inits, setups):
        if self.adopt_point.attached_to:
            return self.adopt_point.attached_to.basic_code(inits, setups)
        else:
            return '0'


class VariableEntryBlock(canvas.Block, XMLBlock):
    def __init__(self, x, y):
        super(VariableEntryBlock, self).__init__(x, y)
        self.text = canvas.Text(0, 0, '0') # Used only for size
        self.outline = canvas.Rectangle(0, 0, 0, 0, 0, fill=(0xFF,0xFF,0xFF,0xFF), stroke=(0xFF,0xFF,0xFF,0xFF))
        self.add(self.outline)
        self.adopt_point = canvas.AttachPoint(0, 0, ['Variable'], self)
        self.adopt_points = [self.adopt_point]
        self.render()

    def layout(self):
        box = self.text.box
        radius = box.height / 2
        self.text.x = radius
        self.text.y = 0
        self.outline.radius = radius
        self.outline.width = box.width + 2 * radius
        self.outline.height = box.height
        if self.adopt_point.attached_to:
            self.outline.visible = False
            self.text.visible = False
            width, height = self.distribute_hozontally(self.children[1:])
        else:
            self.outline.visible = True
            self.text.visible = True
            width, height = self.distribute_hozontally([self.outline])
        self.adopt_point.y = height/2

    def basic_code(self, inits, setups):
        if self.adopt_point.attached_to:
            return self.adopt_point.attached_to.basic_code(inits, setups)
        else:
            return ''


class ValueEntryCtrl(wx.TextCtrl):
    def __init__(self, block, canvas, position, value=0):
        super(ValueEntryCtrl, self).__init__(canvas, wx.ID_ANY, pos=position.Get(), style=wx.TE_PROCESS_ENTER)
        self.Value = str(value)
        self.block = block
        self.SetFocus()
        self.SelectAll()
        self.Bind(wx.EVT_KEY_DOWN, self.on_key_down)
        self.Bind(wx.EVT_KILL_FOCUS, self.on_kill_focus)    # Only seems to be called on OSX when canvas is clicked

    def kerpoof(self, transfer=True):
        if transfer:
            try:
                f = float(self.Value)
            except ValueError:
                pass
            else:
                self.block.value = f
        self.Destroy()

    def on_key_down(self, event):
        keycode = event.GetKeyCode()
        if keycode == wx.WXK_RETURN or keycode == wx.WXK_NUMPAD_ENTER:
            self.kerpoof()
        elif keycode == wx.WXK_ESCAPE:
            self.kerpoof(transfer=False)
        else:
            event.Skip()

    def on_kill_focus(self, event):
        self.kerpoof()


class ValueEntryBlock(canvas.Block, XMLBlock):
    "Value data entry. Can accept value blocks or click to type data."
    def __init__(self, x, y, value, min_value=-2**15, max_value=2**15-1, precision=0):
        super(ValueEntryBlock, self).__init__(x, y)
        self.index = 0
        self.precision = precision
        self.min_value = min_value
        self.max_value = max_value
        self.text = canvas.Text(0, 0, '')#str(value))
        self.outline = canvas.Rectangle(0, 0, 0, 0, 0, fill=(255,255,255,255), stroke=(255,255,255,255))
        self.add(self.outline)
        self.add(self.text)
        self.adopt_point = canvas.AttachPoint(0, 0, ['Value', 'Variable'], self)
        self.adopt_points = [self.adopt_point]
        self.value = value
        self.render()

    @property
    def value(self):
        return self._value

    @value.setter
    def value(self, value):
        if self.precision==0:
            value = int(float(value))
        else:
            value = round(float(value), self.precision)
        if value < self.min_value:
            value = self.min_value
        elif value > self.max_value:
            value = self.max_value
        self._value = value
        self.text.text = str(value)
        self.render()

    def layout(self):
        box = self.text.box
        radius = box.height / 2
        self.text.x = radius
        self.text.y = 0
        self.outline.radius = radius
        self.outline.width = box.width + 2 * radius
        self.outline.height = box.height
        if self.adopt_point.attached_to:
            self.outline.visible = False
            self.text.visible = False
            width, height = self.distribute_hozontally(self.children[2:])
        else:
            self.outline.visible = True
            self.text.visible = True
            width, height = self.distribute_hozontally([self.outline])
        self.adopt_point.y = height/2

    def on_click(self):
        if not self.adopt_point.attached_to:
            ValueEntryCtrl(self, self.canvas, self.canvas.calc_canvas_position(self) + wx.Point2D(0, self.height + 2), self.value)

    def to_xml(self, name=None, position=False):
        element = XMLBlock.to_xml(self, name, position)
        element.text = str(self.value)
        return element

    def from_xml(self, element, canvas_dest):
        if element is not None:
            XMLBlock.from_xml(self, element, canvas_dest)
            self.value = element.text

    def basic_code(self, inits, setups):
        if self.adopt_point.attached_to:
            return self.adopt_point.attached_to.basic_code(inits, setups)
        else:
            return self.value
            #return str(int(self.value))

class StringEntryCtrl(wx.TextCtrl):
    def __init__(self, block, canvas, position, value=''):
        super(StringEntryCtrl, self).__init__(canvas, wx.ID_ANY, pos=position.Get(), style=wx.TE_PROCESS_ENTER)
        self.Value = value
        self.block = block
        self.SetFocus()
        self.SelectAll()
        self.Bind(wx.EVT_KEY_DOWN, self.on_key_down)
        self.Bind(wx.EVT_KILL_FOCUS, self.on_kill_focus)    # Only seems to be called on OSX when canvas is clicked

    def kerpoof(self, transfer=True):
        if transfer:
            self.block.value = self.Value
        self.Destroy()

    def on_key_down(self, event):
        keycode = event.GetKeyCode()
        if keycode == wx.WXK_RETURN or keycode == wx.WXK_NUMPAD_ENTER:
            self.kerpoof()
        elif keycode == wx.WXK_ESCAPE:
            self.kerpoof(transfer=False)
        else:
            event.Skip()

    def on_kill_focus(self, event):
        self.kerpoof()


class StringEntryBlock(canvas.Block, XMLBlock):
    "String data entry."
    def __init__(self, x, y, value=''):
        super(StringEntryBlock, self).__init__(x, y)
        self.text = canvas.Text(0, 0, '')#str(value))
        self.outline = canvas.Rectangle(0, 0, 0, 0, 0, fill=(255,255,255,255), stroke=(255,255,255,255))
        self.add(self.outline)
        self.add(self.text)
        self.adopt_point = canvas.AttachPoint(0, 0, ['String'], self)
        self.adopt_points = [self.adopt_point]
        self.value = value
        self.render()

    @property
    def value(self):
        return self._value

    @value.setter
    def value(self, value):
        self._value = value
        self.text.text = '"' + value + '"'
        self.render()

    def layout(self):
        box = self.text.box
        radius = box.height / 2
        self.text.x = radius
        self.text.y = 0
        self.outline.radius = radius
        self.outline.width = box.width + 2 * radius
        self.outline.height = box.height

    def on_click(self):
        self.test = StringEntryCtrl(self, self.canvas, self.canvas.calc_canvas_position(self) + wx.Point2D(0, self.height + 2), self.value)

    def to_xml(self, name=None, position=False):
        element = XMLBlock.to_xml(self, name, position)
        element.text = self.value
        return element

    def from_xml(self, element, canvas_dest):
        if element is not None:
            XMLBlock.from_xml(self, element, canvas_dest)
            self.value = element.text


class ColourEntryBlock(canvas.Block, XMLBlock):
    "Colour entry block that shows colour_entry picker on click."
    def __init__(self, x, y, colour=(0xFF,0x00,0x00,0xFF)):
        super(ColourEntryBlock, self).__init__(x, y)
        self.colour = colour
        self.outline = canvas.Rectangle(0, 0, 0, 0, 0, fill=(255,255,255,255), stroke=(255,255,255,255))
        self.add(self.outline)
        self.space = canvas.Space(0, 0, 16, 16)
        self.add(self.space)
        self.render()

    def layout(self):
        box = self.space.box
        self.outline.width = box.width
        self.outline.height = box.height
        self.outline.fill = self.colour

    def on_click(self):
        colour = wx.GetColourFromUser(self.canvas, (self.colour))
        if colour.IsOk():
            self.colour = colour
        self.render()

    def to_xml(self, name=None, position=False):
        element = XMLBlock.to_xml(self, name, position)
        ET.SubElement(element, 'red').text = str(self.colour[0])
        ET.SubElement(element, 'green').text = str(self.colour[1])
        ET.SubElement(element, 'blue').text = str(self.colour[2])
        return element

    def from_xml(self, element, canvas_dest):
        if element is not None:
            XMLBlock.from_xml(self, element, canvas_dest)
            red = int(element.find('red').text)
            green = int(element.find('green').text)
            blue = int(element.find('blue').text)
            self.colour = (red, green, blue)
            self.render()


class VarNameEntryCtrl(wx.TextCtrl):
    def __init__(self, canvas, position, var_name=''):
        super(VarNameEntryCtrl, self).__init__(canvas, wx.ID_ANY, pos=position.Get(), style=wx.TE_PROCESS_ENTER)
        self.restoring = False
        if self.check_name(var_name) and var_name != '':
            self.Value = var_name
        else:
            self.Value = 'NewVar'
        self.SetFocus()
        self.SelectAll()
        self.old_name = var_name
        self.Bind(wx.EVT_TEXT, self.on_text)
        self.Bind(wx.EVT_KEY_DOWN, self.on_key_down)
        self.Bind(wx.EVT_KILL_FOCUS, self.on_kill_focus)    # Only seems to be called on OSX when canvas is clicked

    def kerpoof(self, transfer=True):
        if transfer and self.Value != '' and self.check_name(self.Value):
            pub.sendMessage('crumble.variable.rename', old_name=self.old_name, var_name=self.Value)
        self.Destroy()

    def check_name(self, name):
        if name == '':
            return True
        if name[0] not in string.letters:
            return False
        for letter in name:
            if letter not in string.letters and letter != ' ':
                return False
        return True

    def on_text(self, event):
        if self.restoring or self.check_name(self.Value):
            event.Skip()
        else:
            self.restoring = True
            self.Value = self.last_good_value
            self.SetInsertionPoint(self.last_good_insertion_point)
            (start, end) = self.last_good_selection
            self.SetSelection(start, end)
            self.restoring = False
            event.Skip()

    def on_key_down(self, event):
        keycode = event.GetKeyCode()
        if keycode == wx.WXK_RETURN or keycode == wx.WXK_NUMPAD_ENTER:
            self.kerpoof()
        elif keycode == wx.WXK_ESCAPE:
            self.kerpoof(transfer=False)
        else:
            self.last_good_value = self.Value
            self.last_good_insertion_point = self.GetInsertionPoint()
            self.last_good_selection = self.GetSelection()
            event.Skip()

    def on_kill_focus(self, event):
        self.kerpoof()


class WhenBlock(canvas.Block):
    def __init__(self, x, y, text):
        super(WhenBlock, self).__init__(x, y)
        self.outline = canvas.Path('M0.5,16.5036298 C0.5,7.6650691 7.67028489,0.5 16.5013442,0.5 L89.5,0.5 L92.5,3.5 L92.5,16.5 L89.5,19.5 L47.5,19.5 L44.5,22.5 L36.5,22.5 L33.5,19.5 L23.5,19.5 L20.5,22.5 L20.5,28.5 L10.5,38.5 L0.5,28.5 Z')
        self.right_segments = self.outline.path[2:6]
        self.right_reference = self.right_segments[1]
        self.bottom_segments = self.outline.path[4:]
        self.bottom_reference = self.bottom_segments[3]
        self.outline.fill = event_colour
        self.outline.stroke = (0xFF,0xFF,0xFF,0xFF)
        self.outline.stroke_width = 1.0
        self.draggable = True
        self.add(self.outline)
        self.add(canvas.Text(0, 0, text, fill=(0xFF,0xFF,0xFF,0xFF)))
        self.attach_point = canvas.AttachPoint(20, 0, ['Statement'], self)
        self.attach_points = [self.attach_point]
        self.last_attach_point_reversable = True
        self.render()

    def layout(self):
        width, height = self.distribute_hozontally(self.children[1:], top=0, left=0, top_border=6, bottom_border=9, left_border=20, right_border=6, min_height=16, min_width=64)
        self.outline.morph_h(width, self.right_reference, self.right_segments)
        self.outline.morph_v(height, self.bottom_reference, self.bottom_segments)
        self.attach_point.y = height - 3


class StatementBlock(canvas.Block):
    def __init__(self, x, y, fill):
        super(StatementBlock, self).__init__(x, y)
        self.outline = canvas.Path("M0.5,3.5 L3.5,0.5 L13.5,0.5 L16.5,3.5 L24.5,3.5 L27.5,0.5 L77.5,0.5 L80.5,3.5 L80.5,16.5 L77.5,19.5 L27.5,19.5 L24.5,22.5 L16.5,22.5 L13.5,19.5 L3.5,19.5 L0.5,16.5 L0.5,3.5 Z")
        self.right_segments = self.outline.path[6:10]
        self.right_reference = self.right_segments[1]
        self.bottom_segments = self.outline.path[8:]
        self.bottom_reference = self.bottom_segments[3]
        self.outline.fill = fill
        self.outline.stroke = (0xFF,0xFF,0xFF,0xFF)
        self.outline.stroke_width = 1.0
        self.draggable = True
        self.add(self.outline)
        self.attach_offset = wx.Point2D(0, 0)
        self.type_string = 'Statement'
        self.attach_point = canvas.AttachPoint(0, 0, ['Statement'], self)
        self.attach_points = [self.attach_point]
        self.last_attach_point_reversable = True

    def layout(self):
        width, height = self.distribute_hozontally(self.children[1:], top=0, left=0, top_border=3, bottom_border=6, left_border=6, right_border=6, min_height=16, min_width=64)
        self.outline.morph_h(width, self.right_reference, self.right_segments)
        self.outline.morph_v(height, self.bottom_reference, self.bottom_segments)
        self.attach_point.y = height - 3


class IOStatementBlock(StatementBlock):
    def __init__(self, x, y):
        super(IOStatementBlock, self).__init__(x, y, io_colour)


class CrumbStatementBlock(StatementBlock):
    def __init__(self, x, y):
        super(CrumbStatementBlock, self).__init__(x, y, crumb_colour)


class SparkleStatementBlcok(StatementBlock):
    def __init__(self, x, y):
        super(SparkleStatementBlcok, self).__init__(x, y, sparkle_colour)


class ControlStatementBlock(StatementBlock):
    def __init__(self, x, y):
        super(ControlStatementBlock, self).__init__(x, y, control_colour)


class DataStatementBlock(StatementBlock):
    def __init__(self, x, y):
        super(DataStatementBlock, self).__init__(x, y, data_colour)


class MusicStatementBlcok(StatementBlock):
    def __init__(self, x, y):
        super(MusicStatementBlcok, self).__init__(x, y, music_colour)

class ControlBlock(canvas.Block):
    def __init__(self, x, y):
        super(ControlBlock, self).__init__(x, y)
        self.outline = canvas.Path("M0.5,3.5 L3.5,0.5 L13.5,0.5 L16.5,3.5 L24.5,3.5 L27.5,0.5 L77.5,0.5 L80.5,3.5 L80.5,16.5 L77.5,19.5 L39.5,19.5 L36.5,22.5 L28.5,22.5 L25.5,19.5 L15.0727607,19.6755168 L12.5,22.5 L12.5,35.5 L15.5,38.5 L25.5,38.5 L28.5,41.5 L36.5,41.5 L39.5,38.5 L77.5,38.5 L80.5,41.5 L80.5,54.5 L77.5,57.5 L27.5,57.5 L24.5,60.5 L16.5,60.5 L13.5,57.5 L3.5,57.5 L0.5,54.5 Z")
        self.right_segments = self.outline.path[6:10] + self.outline.path[22:26]
        self.right_reference = self.right_segments[1]
        self.bottom_one_segments = self.outline.path[8:]
        self.bottom_one_reference = self.bottom_one_segments[3]
        self.bottom_two_segments = self.outline.path[16:]
        self.bottom_two_reference = self.bottom_two_segments[1]
        self.bottom_three_segments = self.outline.path[24:]
        self.bottom_three_reference = self.bottom_three_segments[3]
        self.outline.fill = control_colour
        self.outline.stroke = (0xFF,0xFF,0xFF,0xFF)
        self.outline.stroke_width = 1.0
        self.add(self.outline)
        self.draggable = True
        self.attach_offset = wx.Point2D(0, 0)
        self.type_string = 'Statement'
        self.adopt_point = canvas.AttachPoint(12, 0, ['Statement'], self)
        self.adopt_points = [self.adopt_point]
        self.attach_point = canvas.AttachPoint(0, 0, ['Statement'], self)
        self.attach_points = [self.attach_point]
        self.last_attach_point_reversable = True
        self.render()

    def layout(self):
        y = 0
        top_children = [c for c in self.children if c.layout_hint=='top']
        width, height = self.distribute_hozontally(top_children, top=y, left=0, top_border=3, bottom_border=6, left_border=6, right_border=6, min_height=16, min_width=64)
        self.outline.morph_h(width, self.right_reference, self.right_segments)
        self.outline.morph_v(height, self.bottom_one_reference, self.bottom_one_segments)
        y += height - 3
        self.adopt_point.y = y
        if self.adopt_point.attached_to:
            block = self.adopt_point.attached_to
            self.canvas.move_all_attached(block, self.adopt_point - block.attach_offset)
            attached = self.canvas.find_all_attached(block)
            y = attached[-1].position.y + attached[-1].height - 3
        else:
            y+=12
        self.outline.morph_v(y, self.bottom_two_reference, self.bottom_two_segments)
        bottom_children = [c for c in self.children if c.layout_hint=='bottom']
        width, height = self.distribute_hozontally(bottom_children, top=y, left=0, top_border=3, bottom_border=6, left_border=6, right_border=6, min_height=16, min_width=64)
        y += height
        self.outline.morph_v(y, self.bottom_three_reference, self.bottom_three_segments)
        self.attach_point.y = y - 3


class ControlDoubleBlock(canvas.Block):
    def __init__(self, x, y):
        super(ControlDoubleBlock, self).__init__(x, y)
        self.outline = canvas.Path("M0.5,3.5 L3.5,0.5 L13.5,0.5 L16.5,3.5 L24.5,3.5 L27.5,0.5 L77.5,0.5 L80.5,3.5 L80.5,16.5 L77.5,19.5 L39.5,19.5 L36.5,22.5 L28.5,22.5 L25.5,19.5 L15.0727607,19.6755168 L12.5,22.5 L12.5,35.5 L15.5,38.5 L25.5,38.5 L28.5,41.5 L36.5,41.5 L39.5,38.5 L77.5,38.5 L80.5,41.5 L80.5,54.5 L77.5,57.5 L39.5,57.5 L36.5,60.5 L28.5,60.5 L25.5,57.5 L15.5,57.5 L12.5,60.5 L12.5,73.5 L15.5,76.5 L25.5,76.5 L28.5,79.5 L36.5,79.5 L39.5,76.5 L77.5,76.5 L80.5,79.5 L80.5,92.5 L77.5,95.5 L27.5,95.5 L24.5,98.5 L16.5,98.5 L13.5,95.5 L3.5,95.5 L0.5,92.5 Z")
        self.right_segments = self.outline.path[6:10] + self.outline.path[22:26] + self.outline.path[38:42]
        self.right_reference = self.right_segments[1]
        self.bottom_one_segments = self.outline.path[8:]
        self.bottom_one_reference = self.bottom_one_segments[3]
        self.bottom_two_segments = self.outline.path[16:]
        self.bottom_two_reference = self.bottom_two_segments[1]
        self.bottom_three_segments = self.outline.path[24:]
        self.bottom_three_reference = self.bottom_three_segments[3]
        self.bottom_four_segments = self.outline.path[32:]
        self.bottom_four_reference = self.bottom_four_segments[1]
        self.bottom_five_segments = self.outline.path[40:]
        self.bottom_five_reference = self.bottom_five_segments[3]
        self.outline.fill = control_colour
        self.outline.stroke = (0xFF,0xFF,0xFF,0xFF)
        self.outline.stroke_width = 1.0
        self.add(self.outline)
        self.draggable = True
        self.attach_offset = wx.Point2D(0, 0)
        self.type_string = 'Statement'
        self.adopt_point_one = canvas.AttachPoint(12, 0, ['Statement'], self)
        self.adopt_point_two = canvas.AttachPoint(12, 0, ['Statement'], self)
        self.adopt_points = [self.adopt_point_one, self.adopt_point_two]
        self.attach_point = canvas.AttachPoint(0, 0, ['Statement'], self)
        self.attach_points = [self.attach_point]
        self.last_attach_point_reversable = True
        self.render()

    def layout(self):
        y = 0
        top_children = [c for c in self.children if c.layout_hint=='top']
        width, height = self.distribute_hozontally(top_children, top=y, left=0, top_border=3, bottom_border=6, left_border=6, right_border=6, min_height=16, min_width=64)
        self.outline.morph_h(width, self.right_reference, self.right_segments)
        self.outline.morph_v(height, self.bottom_one_reference, self.bottom_one_segments)
        y += height - 3
        self.adopt_point_one.y = y
        if self.adopt_point_one.attached_to:
            block = self.adopt_point_one.attached_to
            self.canvas.move_all_attached(block, self.adopt_point_one - block.attach_offset)
            attached = self.canvas.find_all_attached(block)
            y = attached[-1].position.y + attached[-1].height - 3
        else:
            y+=12
        self.outline.morph_v(y, self.bottom_two_reference, self.bottom_two_segments)
        middle_children = [c for c in self.children if c.layout_hint=='middle']
        width, height = self.distribute_hozontally(middle_children, top=y, left=0, top_border=3, bottom_border=6, left_border=6, right_border=6, min_height=16, min_width=64)
        y += height
        self.outline.morph_v(y, self.bottom_three_reference, self.bottom_three_segments)
        self.adopt_point_two.y = y - 3
        if self.adopt_point_two.attached_to:
            block = self.adopt_point_two.attached_to
            self.canvas.move_all_attached(block, self.adopt_point_two - block.attach_offset)
            attached = self.canvas.find_all_attached(block)
            y = attached[-1].position.y + attached[-1].height - 3
        else:
            y+=12
        self.outline.morph_v(y, self.bottom_four_reference, self.bottom_four_segments)
        bottom_children = [c for c in self.children if c.layout_hint=='bottom']
        width, height = self.distribute_hozontally(bottom_children, top=y, left=0, top_border=3, bottom_border=6, left_border=6, right_border=6, min_height=16, min_width=64)
        y += height
        self.outline.morph_v(y, self.bottom_five_reference, self.bottom_five_segments)
        self.attach_point.y = y - 3


class ControlForeverBlock(canvas.Block):
    def __init__(self, x, y):
        super(ControlForeverBlock, self).__init__(x, y)
        self.outline = canvas.Path("M0.5,3.5 L3.5,0.5 L13.5,0.5 L16.5,3.5 L24.5,3.5 L27.5,0.5 L77.5,0.5 L80.5,3.5 L80.5,16.5 L77.5,19.5 L39.5,19.5 L36.5,22.5 L28.5,22.5 L25.5,19.5 L15.0727607,19.6755168 L12.5,22.5 L12.5,35.5 L15.5,38.5 L25.5,38.5 L28.5,41.5 L36.5,41.5 L39.5,38.5 L77.5,38.5 L80.5,41.5 L80.5,54.5 L77.5,57.5 L3.5,57.5 L0.5,54.5 Z")
        self.right_segments = self.outline.path[6:10] + self.outline.path[22:26]
        self.right_reference = self.right_segments[1]
        self.bottom_one_segments = self.outline.path[8:]
        self.bottom_one_reference = self.bottom_one_segments[3]
        self.bottom_two_segments = self.outline.path[16:]
        self.bottom_two_reference = self.bottom_two_segments[1]
        self.bottom_three_segments = self.outline.path[24:]
        self.bottom_three_reference = self.bottom_three_segments[1]
        self.outline.fill = control_colour
        self.outline.stroke = (0xFF,0xFF,0xFF,0xFF)
        self.outline.stroke_width = 1.0
        self.add(self.outline)
        self.bottom_spacer = canvas.Space(12, 0, 3, 3)
        self.add(self.bottom_spacer)
        self.draggable = True
        self.attach_offset = wx.Point2D(0, 0)
        self.type_string = 'Statement'
        self.adopt_point = canvas.AttachPoint(12, 0, ['Statement'], self)
        self.adopt_points = [self.adopt_point]
        self.render()

    def layout(self):
        y = 0
        top_children = [c for c in self.children if c.layout_hint=='top']
        width, height = self.distribute_hozontally(top_children, top=y, left=0, top_border=3, bottom_border=6, left_border=6, right_border=6, min_height=16, min_width=64)
        self.outline.morph_h(width, self.right_reference, self.right_segments)
        self.outline.morph_v(height, self.bottom_one_reference, self.bottom_one_segments)
        y += height - 3
        self.adopt_point.y = y
        if self.adopt_point.attached_to:
            block = self.adopt_point.attached_to
            self.canvas.move_all_attached(block, self.adopt_point - block.attach_offset)
            attached = self.canvas.find_all_attached(block)
            y = attached[-1].position.y + attached[-1].height - 3
        else:
            y+=12
        self.outline.morph_v(y, self.bottom_two_reference, self.bottom_two_segments)
        bottom_children = [c for c in self.children if c.layout_hint=='bottom']
        width, height = self.distribute_hozontally(bottom_children, top=y, left=0, top_border=3, bottom_border=3, left_border=6, right_border=6, min_height=16, min_width=64)
        y += height
        self.outline.morph_v(y, self.bottom_three_reference, self.bottom_three_segments)
        self.bottom_spacer.y = y


class BooleanBlock(canvas.Block):
    def __init__(self, x, y):
        super(BooleanBlock, self).__init__(x, y)
        self.outline = canvas.RightAngledHexagon(0, 0, 0, 0, fill=(0x4C,0xD9,0x64,255), stroke=(255,255,255,255), stroke_width=1.0)
        self.draggable = True
        self.add(self.outline)
        self.attach_offset = wx.Point2D(0, 0)
        self.type_string = 'Bool'

    def layout(self):
        width, height = self.distribute_hozontally(self.children[1:], top_border=3, bottom_border=3,
                                        left_border=10, right_border=10, min_height=10, min_width=32)
        self.outline.width = width
        self.outline.height = height
        self.attach_offset = wx.Point2D(0,height/2)


class ValueBlock(canvas.Block):
    def __init__(self, x, y):
        super(ValueBlock, self).__init__(x, y)
        self.outline = canvas.Rectangle(0, 0, 0, 0, 0, fill=(0xC8,0x6E,0xDF,255), stroke=(255,255,255,255))
        self.draggable = True
        self.add(self.outline)
        self.attach_offset = wx.Point2D(0, 0)
        self.type_string = 'Value'

    def layout(self):
        width, height = self.distribute_hozontally(self.children[1:], top_border=3, bottom_border=3,
                                        left_border=10, right_border=10, min_height=10, min_width=20)
        self.outline.width = width
        self.outline.height = height
        self.outline.radius = height/2
        self.attach_offset = wx.Point2D(0,height/2)


class MonitorBlock(canvas.Block):
    def __init__(self, x, y):
        super(MonitorBlock, self).__init__(x, y)
        self.outline = canvas.Rectangle(0, 0, 0, 0, 0, fill=(0xFF,0xFF,0xFF,0xFF), stroke=(0x00,0x00,0x00,0XFF))
        self.draggable = False
        self.add(self.outline)
        self.type_string = 'Monitor'

    def layout(self):
        width, height = self.distribute_hozontally(self.children[1:], top_border=3, bottom_border=3,
                                        left_border=10, right_border=10, min_height=10, min_width=20)
        self.outline.width = width
        self.outline.height = height
        self.outline.radius = height/2


class OperatorBlock(canvas.Block):
    def __init__(self, x, y):
        super(OperatorBlock, self).__init__(x, y)
        self.outline = canvas.Rectangle(0, 0, 0, 0, 0, fill=operator_colour, stroke=(255,255,255,255))
        self.draggable = True
        self.add(self.outline)
        self.attach_offset = wx.Point2D(0, 0)
        self.type_string = 'Value'

    def layout(self):
        width, height = self.distribute_hozontally(self.children[1:], top_border=3, bottom_border=3,
                                        left_border=4, right_border=4, min_height=10, min_width=32)
        self.outline.width = width
        self.outline.height = height
        self.outline.radius = height/2
        self.attach_offset = wx.Point2D(0,height/2)

