# -*- coding: utf-8 -*-

#############################################################################
# Copyright (c) 2014-2017 Redfern Electronics Limited. All rights reserved. #
#############################################################################

import numpy as np
import wx
from pyparsing import Word, Suppress, Literal, OneOrMore, nums
from lang import _
import lang

try:
    from wx.lib.graphics import GraphicsContext, GraphicsPath, GraphicsBitmap
except ImportError:
    GraphicsContext = wx.GraphicsContext
    GraphicsPath = wx.GraphicsPath

class Canvas(wx.Panel):
    """children used when members can be to shapes and blocks. blocks used when members will only be blocks."""
    def __init__(self, parent, id):
        super(Canvas, self).__init__(parent, id)
        self.SetBackgroundColour(wx.WHITE)
        self.children = []
        self.floating_blocks = []
        self.markers = []
        self.arrow_marker = Marker(0, 0)
        self.add(self.arrow_marker, self.markers)
        self.toolboxes = []
        self.has_moved = False
        self.draggable_hit = None
        self.toolbox_hit = None
        self.attach_distance_squared = 32**2
        self.panning = False
        self.last_position = None
        self.modified = False

        self.Bind(wx.EVT_MOTION, self.on_motion)
        self.Bind(wx.EVT_LEFT_DOWN, self.on_left_down)
        self.Bind(wx.EVT_LEFT_UP, self.on_left_up)
        self.Bind(wx.EVT_RIGHT_DOWN, self.on_right_down)
        self.Bind(wx.EVT_PAINT, self.on_paint)
        self.Bind(wx.EVT_ERASE_BACKGROUND, self.on_erase)
        self.Bind(wx.EVT_MOUSE_CAPTURE_LOST, self.on_mouse_capture_lost)

    def add(self, new_child, dest=None):
        if dest is None:
            dest = self.children
        new_child.parent = self
        new_child.canvas = self
        dest.append(new_child)

    def remove(self, child):
        child.parent = None
        child.canvas = None
        self.children.remove(child)

    def delete_all(self):
        del self.children[:]
        self.Refresh()

    def find_draggable(self, children, point):
        """Searches recursively through children for highest level draggable block."""
        for child in reversed(children):
            if hasattr(child, 'contains') and child.contains(point):
                draggable = self.find_draggable(child.children, point - child.position)
                if draggable:
                    return draggable
                if child.draggable:
                    return child

    def find_clickable(self, children, point):
        """Searches recursively through children for highest level non-draggable block."""
        for child in reversed(children):
            if hasattr(child, 'contains') and child.contains(point):
                clickable = self.find_clickable(child.children, point - child.position)
                if clickable:
                    return clickable
                if not child.draggable:
                    return child

    def detach(self, block):
        """Detach a block (slave) from a attach point."""
        if block.attached_to is not None:
            block.attached_to.attached_to = None
            block.attached_to = None

    def attach_blocks_to_point(self, attach_point):
        """Attaches first floating block to the attach point. If point's
        parent block's parent is not the canvas floating blocks will be adopted.
        Will attempt to re-attach existing block to last block
        if types match. Blocks need to be attached apart from last block.
        Moves blocks. Returns a redraw rect if parent is canvas, otherwise
        the adopting parent is rendered."""
        old_block = attach_point.attached_to
        if old_block:
            self.detach(old_block)
        first_block = self.floating_blocks[0]
        last_block = self.floating_blocks[-1]
        self.unfloat(attach_point.parent.parent)
        attach_point.attached_to = first_block
        first_block.attached_to = attach_point
        rect = self.move_all_attached(first_block, attach_point.parent.position
                                      + attach_point - first_block.attach_offset)
        if (old_block and last_block.attach_points
             and old_block.type_string in last_block.attach_points[0].type_strings):
            last_block.attach_points[0].attached_to = old_block
            old_block.attached_to = last_block.attach_points[0]
            rect.Union(self.move_all_attached(old_block, last_block.position
                                              + last_block.attach_points[0]
                                              - old_block.attach_offset))
        elif old_block:
            self.float_all_attached(old_block)
            rect.Union(self.move_all_attached(old_block, old_block.position + wx.Point2D(first_block.box.width + 8, 8)))
            self.unfloat(self)

        if attach_point.parent.parent is not self:
            attach_point.parent.parent.render()
        else:
            return rect

    def adopt_floating_to_point(self, adopt_point):
        """Attaches first floating block to the adopt point.
        Floating blocks will be adopted.
        Will attempt to re-attach existing block to last block
        if types match. If can't reattach offset blocks to canvas.
        Blocks need to be attached apart from last block.
        Moves blocks. Renders adopting parent."""
        old_block = adopt_point.attached_to
        if old_block:
            self.detach(old_block)
        first_block = self.floating_blocks[0]
        last_block = self.floating_blocks[-1]
        self.unfloat(adopt_point.parent)
        adopt_point.attached_to = first_block
        first_block.attached_to = adopt_point
        rect = self.move_all_attached(first_block, adopt_point - first_block.attach_offset)
        if (old_block and last_block.attach_points
             and old_block.type_string in last_block.attach_points[0].type_strings):
            last_block.attach_points[0].attached_to = old_block
            old_block.attached_to = last_block.attach_points[0]
            rect.Union(self.move_all_attached(old_block, last_block.position
                                              + last_block.attach_points[0]
                                              - old_block.attach_offset))
        elif old_block:
            self.float_all_attached(old_block)
            self.move_all_attached(old_block, old_block.position + wx.Point2D(first_block.box.width + 8, 8))
            self.unfloat(self)
        adopt_point.parent.render()

    def attach_floating_to_block(self, block):
        """Attaches the last floating block's attach point to block.
        Should only be used if block is unattached.
        Returns a redraw rect."""
        old_attach_point = block.attached_to
        first_block = self.floating_blocks[0]
        last_block = self.floating_blocks[-1]
        last_block.attach_points[0].attached_to = block
        block.attached_to = last_block.attach_points[0]
        rect = self.move_all_attached(first_block, block.position + block.attach_offset
                                          - (last_block.position
                                             + last_block.attach_points[0]
                                             - first_block.position))
        self.unfloat(self)
        return rect

    def delete_floating_blocks(self):
        first_block = self.floating_blocks[0]
        rect = self.move_all_attached(first_block, first_block.position)
        del self.floating_blocks[:]
        return rect

    def find_all_attached(self, block):
        """Returns a list of all blocks attached to and including block."""
        blocks = [block]
        for attach_point in block.attach_points:
            if attach_point.attached_to:
                blocks += self.find_all_attached(attach_point.attached_to)
        return blocks

    def calc_canvas_position(self, block):
        """Return a wx.Point2D representing the position of a block/subblock on the canvas."""
        if block.parent==self:
            return wx.Point2DCopy(block.position)
        else:
            return self.calc_canvas_position(block.parent) + block.position

    def float_all_attached(self, block):
        """Move block and all attached blocks to the floating layer."""
        blocks = self.find_all_attached(block)
        if block.parent is not self:
            render_parent = block.parent
        else:
            render_parent = None
        for block in blocks:
            canvas_position = self.calc_canvas_position(block)
            block.parent.children.remove(block)
            self.floating_blocks.append(block)
            block.parent = self
            block.position = canvas_position
        if render_parent:
            render_parent.render()

    def unfloat(self, parent):
        """Move floating blocks parent's children."""
        for block in self.floating_blocks:
            parent.add(block)
        del self.floating_blocks[:]

    def find_attachable_blocks(self, children, attach_point):
        """Call this (non-recursive) function with all canvas children
        to get a list of matching attachable blocks (no sub-blocks) that
        are not currently attached. And a list of their attach offset
        coords on the canvas."""
        blocks = []
        coords = []
        for child in children:
            if child.attach_offset is not None and child.type_string in attach_point.type_strings and not child.attached_to:
                blocks.append(child)
                coords.append(child.position + child.attach_offset)
        return blocks, coords

    def find_attach_points(self, children, type_string):
        """Call this recursive function with all canvas children
        to get a list of matching attach points and a list
        of their coords on the canvas."""
        points = []
        coords = []
        for child in children:
            if hasattr(child, 'attach_points'):
                for attach_point in child.attach_points:
                    if type_string in attach_point.type_strings:
                        points.append(attach_point)
                        coords.append(attach_point + self.calc_canvas_position(child))
                if hasattr(child, 'children'):
                    p, c = self.find_attach_points(child.children, type_string)
                    points += p
                    coords += c

        return points, coords

    def find_adopt_points(self, children, type_string):
        """Call this recursive function with all canvas children
        to get a list of matching adopt points and a list
        of their coords on the canvas."""
        points = []
        coords = []
        for child in children:
            if hasattr(child, 'adopt_points'):
                for adopt_point in child.adopt_points:
                    if type_string in adopt_point.type_strings:
                        points.append(adopt_point)
                        coords.append(adopt_point + self.calc_canvas_position(child))
                if hasattr(child, 'children'):
                    p, c = self.find_adopt_points(child.children, type_string)
                    points += p
                    coords += c
        return points, coords

    def find_closest_attachble_block(self, max_distance_squared):
        """Finds the closest attachable block to the last floating block."""
        if self.floating_blocks:
            last_block = self.floating_blocks[-1]
            if self.attachable_block_coords.size:
                distances_squared = np.sum((self.attachable_block_coords - np.array([last_block.attach_points[-1] + last_block.position]))**2, axis=1)
                index_of_closest = np.argmin(distances_squared)
                closest = self.attachable_blocks[index_of_closest]
                distance_squared = distances_squared[index_of_closest]

                if distance_squared < max_distance_squared:
                    return closest, index_of_closest, distance_squared
        return None, None, max_distance_squared

    def find_closest_attach_point(self, max_distance_squared):
        """Finds the closest attach point to the first floating block."""
        if self.floating_blocks:
            first_block = self.floating_blocks[0]
            if self.attach_coords.size:
                distances_squared = np.sum((self.attach_coords - np.array([first_block.attach_offset + self.calc_canvas_position(first_block)]))**2, axis=1)
                index_of_closest = np.argmin(distances_squared)
                closest = self.attach_points[index_of_closest]
                distance_squared = distances_squared[index_of_closest]

                if distance_squared < max_distance_squared:
                    return closest, index_of_closest, distance_squared
        return None, None, max_distance_squared

    def find_closest_adopt_point(self, max_distance_squared):
        """Finds the closest adopt point to the first floating block."""
        if self.floating_blocks:
            first_block = self.floating_blocks[0]
            if self.adopt_coords.size:
                distances_squared = np.sum((self.adopt_coords - np.array([first_block.attach_offset + self.calc_canvas_position(first_block)]))**2, axis=1)
                index_of_closest = np.argmin(distances_squared)
                closest = self.adopt_points[index_of_closest]
                distance_squared = distances_squared[index_of_closest]

                if distance_squared < max_distance_squared:
                    return closest, index_of_closest, distance_squared
        return None, None, max_distance_squared

    def move_with_offset(self, blocks, position):
        """Moves block to position + their hit offset and return rectangle to redraw.
        Faster than using move_all_attached when dragging."""
        if blocks:
            rect = blocks[0].box
        for block in blocks:
            rect.Union(block.box)
            block.position = position + block.hit_offset
            rect.Union(block.box)
        return rect

    def move_all_attached(self, block, position):
        """Move block to position and move all attached block with it."""
        rect = block.box
        block.position = position
        rect.Union(block.box)
        for attach_point in block.attach_points:
            if attach_point.attached_to:
                rect.Union(self.move_all_attached(attach_point.attached_to,
                                                  block.position + attach_point
                                                  - attach_point.attached_to.attach_offset))
        return rect

    def on_duplicate(self, event):
        self.modified = True
        element = self.draggable_hit.to_xml(position=True)
        import blocks
        first_block = blocks.from_xml(element, self)
        # self.float_all_attached(first_block)
        self.render(self.move_all_attached(first_block, self.calc_canvas_position(self.draggable_hit) + wx.Point2D(self.draggable_hit.box.width + 8, 8)))
        # self.unfloat(self)

    def on_motion(self, event):
        position = wx.Point2DFromPoint(event.GetPosition())
        # Check for bogus motion event (e.g. when using VNC)
        if self.last_position == position:
            return
        else:
            self.last_position = position

        if self.panning:
            rect = self.move_with_offset(self.children, position)
            self.render(rect)
        elif self.toolbox_hit in self.toolboxes:
            x = 0
            y = position.y + self.toolbox_hit.hit_offset.y
            if y > 0 or self.Rect.Height > self.toolbox_hit.box.height:
                y = 0
            elif y + self.toolbox_hit.box.height < self.Size[1]:
                y = self.Size[1] - self.toolbox_hit.box.height
            rect = self.toolbox_hit.move(wx.Point2D(x, y))
            self.render(rect)
        elif not self.has_moved and (self.draggable_hit or self.toolbox_hit):
            if self.draggable_hit:
                self.detach(self.draggable_hit)
                self.float_all_attached(self.draggable_hit)
            elif self.toolbox_hit:
                new_block = self.toolbox_hit.spawn()
                self.add(new_block)
                self.float_all_attached(new_block)
            first_block = self.floating_blocks[0]
            last_block = self.floating_blocks[-1]

            if first_block.attach_offset is not None:
                self.attach_points, coords = self.find_attach_points(self.children, first_block.type_string)
                self.attach_coords = np.asarray(coords)
                self.adopt_points, coords = self.find_adopt_points(self.children, first_block.type_string)
                self.adopt_coords = np.asarray(coords)
            else:
                self.attach_points = []
                self.adopt_points = []
                coords = []
                self.attach_coords = np.asarray(coords)
                self.adopt_coords = np.asarray(coords)

            if last_block.attach_points and last_block.last_attach_point_reversable:
                self.attachable_blocks, coords = self.find_attachable_blocks(self.children,
                                                                             last_block.attach_points[-1])
            else:
                self.attachable_blocks = []
                coords = []
            self.attachable_block_coords = np.array(coords)

        self.has_moved = True

        if self.floating_blocks:

            rect = self.move_with_offset(self.floating_blocks, position)

            closest_attach_point, closest_attach_index, distance_squared = self.find_closest_attach_point(self.attach_distance_squared)
            closest_adopt_point, closest_adopt_index, distance_squared = self.find_closest_adopt_point(distance_squared)
            closest_attachable_block, closest_attachable_index, distance_squared = self.find_closest_attachble_block(distance_squared)

            if closest_attachable_block:
                self.arrow_marker.visible = True
                x = self.attachable_block_coords[closest_attachable_index][0]
                y = self.attachable_block_coords[closest_attachable_index][1]
                rect.Union(self.arrow_marker.point_to(wx.Point2D(x, y)))
            elif closest_adopt_point:
                self.arrow_marker.visible = True
                x = self.adopt_coords[closest_adopt_index][0]
                y = self.adopt_coords[closest_adopt_index][1]
                rect.Union(self.arrow_marker.point_to(wx.Point2D(x, y)))
            elif closest_attach_point:
                self.arrow_marker.visible = True
                x = self.attach_coords[closest_attach_index][0]
                y = self.attach_coords[closest_attach_index][1]
                rect.Union(self.arrow_marker.point_to(wx.Point2D(x, y)))
            else:
                if self.arrow_marker.visible==True:
                    self.arrow_marker.visible = False
                    rect.Union(self.arrow_marker.point_to(self.arrow_marker.position)) # Minimise redraw rect size

            self.render(rect)

    def on_left_down(self, event):
        self.modified = True
        self.CaptureMouse()
        focus = self.FindFocus()
        if focus and focus is not self and hasattr(focus, 'kerpoof'):
            focus.kerpoof()
        position = wx.Point2DFromPoint(event.GetPosition())
        self.draggable_hit = self.find_draggable(self.children, position)
        self.toolbox_hit = self.find_draggable(self.toolboxes, position)
        if self.draggable_hit:
            attached_blocks = self.find_all_attached(self.draggable_hit)
            for block in attached_blocks:
                block.hit_offset = self.calc_canvas_position(block) - position
        elif self.toolbox_hit:
            self.toolbox_hit.hit_offset = self.calc_canvas_position(self.toolbox_hit) - position
        else:
            self.SetCursor(wx.StockCursor(wx.CURSOR_SIZING))
            for child in self.children:
                child.hit_offset = self.calc_canvas_position(child) - position
                self.panning = True
        self.has_moved = False

    def on_left_up(self, event):
        self.draggable_hit = None
        self.toolbox_hit = None
        self.panning = False

        if self.GetCapture() == self:
            self.ReleaseMouse()
        self.SetCursor(wx.StockCursor(wx.CURSOR_DEFAULT))
        if self.has_moved and self.floating_blocks:
            if self.toolboxes and self.floating_blocks[0].box.Intersects(self.toolboxes[0].box):
                self.render(self.delete_floating_blocks())
            else:
                closest_attach_point, index, distance_squared = self.find_closest_attach_point(self.attach_distance_squared)
                closest_adopt_point, index, distance_squared = self.find_closest_adopt_point(distance_squared)
                closest_attachable_block, index, distance_squared = self.find_closest_attachble_block(distance_squared)
                if closest_attachable_block:
                    self.render(self.attach_floating_to_block(closest_attachable_block))
                elif closest_adopt_point:
                    self.render(self.adopt_floating_to_point(closest_adopt_point))
                elif closest_attach_point:
                    self.render(self.attach_blocks_to_point(closest_attach_point))
                else:
                    self.unfloat(self)
        elif not self.has_moved:
            position = wx.Point2DFromPoint(event.GetPosition())
            clickable_hit = self.find_clickable(self.children, position)
            if clickable_hit:
                clickable_hit.on_click()
            toolbox_hit = self.find_clickable(self.toolboxes, position)
            try:
                if toolbox_hit.toolbox_click:
                    toolbox_hit.on_click()
            except AttributeError:
                pass

        self.arrow_marker.visible = False
        self.arrow_marker.render()

    def on_right_down(self, event):
        position = wx.Point2DFromPoint(event.GetPosition())
        self.draggable_hit = self.find_draggable(self.children, position)
        if self.draggable_hit:
            menu = wx.Menu()
            item = wx.MenuItem(menu, wx.NewId(), _("Duplicate"))
            menu.AppendItem(item)
            self.Bind(wx.EVT_MENU, self.on_duplicate, item)
            self.PopupMenu(menu, event.GetPosition())
            menu.Destroy()

    def on_mouse_capture_lost(self, event):
        pass

    def render(self, rect=None):
        if rect:
            rect.Inset(-2.0, -2.0)
            self.RefreshRect(rect=rect.Get(), eraseBackground=False)
        else:
            self.Refresh(eraseBackground=False)

    def on_paint(self, event):
        if wx.Platform == '__WXMSW__':
            dc = wx.BufferedPaintDC(self)
        else:
            dc = wx.PaintDC(self)
        dc.Clear()
        gc = GraphicsContext.Create(dc)
        for child in self.children:
            child.draw(gc)
        for marker in self.markers:
            marker.draw(gc)
        for toolbox in self.toolboxes:
            toolbox.draw(gc)
        for child in self.floating_blocks:
            child.draw(gc)

    def on_erase(self, event):
        pass


class Shape(object):
    "Base class for all objects on the canvas"
    def __init__(self, x, y, width, height, fill=(255,0,0,128), stroke=(0,0,0,255), stroke_width=1.0):
        self.x = x
        self.y = y
        self.width = width
        self.height = height
        self.fill = fill
        self.stroke = stroke
        self.stroke_width = stroke_width
        self.visible = True
        self.parent = None
        self.canvas = None
        self.layout_hint = ''

    @property
    def position(self):
        return wx.Point2D(self.x, self.y)
    @position.setter
    def position(self, p):
        self.x = p.x
        self.y = p.y

    @property
    def box(self):
        return wx.Rect2D(self.position.x, self.position.y, self.width, self.height)

    def draw_shape(self, gc):
        pass

    def draw(self, gc):
        if self.visible:
            gc.PushState()
            gc.SetPen(wx.Pen(self.stroke, self.stroke_width))
            gc.SetBrush(wx.Brush(self.fill))
            self.draw_shape(gc)
            gc.PopState()


class Rectangle(Shape):
    "A rounded rectangle."
    def __init__(self, x, y, width, height, radius, fill=(0,0,0,128), stroke=(0,0,0,255), stroke_width=1.0):
        super(Rectangle, self).__init__(x, y, width, height, fill, stroke, stroke_width)
        self.radius = radius

    def draw_shape(self, gc):
        gc.DrawRoundedRectangle(self.x, self.y, self.width, self.height, self.radius)


class RightAngledHexagon(Shape):
    "A hexagon with two opposite right angles."
    def __init__(self, x, y, width, height, fill=(0,0,0,128), stroke=(0,0,0,255), stroke_width=1.0):
        super(RightAngledHexagon, self).__init__(x, y, width, height, fill, stroke, stroke_width)

    def draw_shape(self, gc):
        path = gc.CreatePath()
        base = self.height / 2.0
        path.MoveToPoint(0, base)
        path.AddLineToPoint(base, 0)
        path.AddLineToPoint(self.width - base, 0)
        path.AddLineToPoint(self.width, base)
        path.AddLineToPoint(self.width - base, self.height)
        path.AddLineToPoint(base, self.height)
        path.CloseSubpath()
        gc.Translate(self.x, self.y)
        gc.DrawPath(path)


class Space(Shape):
    "Space helper. No drawing is done."
    def __init__(self, x, y, width, height):
        super(Space, self).__init__(x, y, width, height)

    def draw(self, gc):
        "Nothing to draw for space."
        pass


class Bitmap(Shape):
    "Draws bitmap from file"
    def __init__(self, x, y, bitmap):
        super(Bitmap, self).__init__(x, y, 0, 0)
        if wx.Platform == '__WXGTK__':
            # Horrible hack to make alpha work on wx.GTK
            # Tried loading RGBA data into numpy buffer and using
            # createfrombuffer. Image appears but where alpha is not 0 or
            # 255 compositing is wrong.
            # TODO: Put file in proper temp dir (or find fix)
            # Might be better to load all images into approprate bitmaps and
            # reference single copy (new png is created every time Bitmap is create
            # E.g. when crumble is spawned from toolbox
            bitmap.SaveFile('temp.png', wx.BITMAP_TYPE_PNG)
            self.bitmap = GraphicsBitmap.CreateFromPNG('temp.png')
        else:
            self.bitmap = bitmap
        self.width = self.bitmap.Width
        self.height = self.bitmap.Height

    def draw_shape(self, gc):
        gc.DrawBitmap(self.bitmap, self.x, self.y, self.width, self.height)


# Path object and path segments

class Point(wx.Point2D):
    def __init__(self, x=0.0, y=0.0):
        super(Point, self).__init__(x, y)

    @classmethod
    def from_svg(cls, s, l, t):
        return cls(t[0]-0.5, t[1]-0.5) # Adjust by half a pixel for SVG differences

class MoveTo(object):
    def __init__(self, point, relative=False):
        self.relative = relative
        self.point = point

    def move_by(self, dx, dy):
        self.point.x += dx
        self.point.y += dy

    def draw(self, graphics_path):
        graphics_path.MoveToPoint(self.point.x, self.point.y)

    @classmethod
    def from_svg(cls, s, l, t):
        return cls(t[1], False if t[0].isupper() else True)

class ClosePath(object):
    def __init__(self):
        pass

    def move_by(self, dx, dy):
        pass

    def draw(self, graphics_path):
        graphics_path.CloseSubpath()

    @classmethod
    def from_svg(cls, s, l, t):
        return cls()

class LineTo(object):
    def __init__(self, point, relative=False):
        self.relative = relative
        self.point = point

    def move_by(self, dx, dy):
        self.point.x += dx
        self.point.y += dy

    def draw(self, graphics_path):
        graphics_path.AddLineToPoint(self.point.x, self.point.y)

    @classmethod
    def from_svg(cls, s, l, t):
        return cls(t[1], False if t[0].isupper() else True)

class CurveTo(object):
    def __init__(self, control_point1, control_point2, point, relative=False):
        self.point = point
        self.control_point1 = control_point1
        self.control_point2 = control_point2
        self.relative = relative

    def move_by(self, dx, dy):
        for p in [self.point, self.control_point1, self.control_point2]:
            p.x += dx
            p.y += dy

    def draw(self, graphics_path):
        graphics_path.AddCurveToPoint(self.control_point1.x, self.control_point1.y,
                                      self.control_point2.x, self.control_point2.y,
                                      self.point.x, self.point.y)

    @classmethod
    def from_svg(cls, s, l, t):
        return cls(t[1], t[2], t[3], False if t[0].isupper() else True)


class Path(object):
    "Path object for complex shapes. Can be initialise with SVG path."
    def __init__(self, svg_path="", fill=(0,0,0,128), stroke=(0,0,0,255), stroke_width=1.0):
        self.fill = fill
        self.stroke = stroke
        self.stroke_width = stroke_width
        if svg_path:
            self.path = self.from_svg(svg_path)
        else:
            self.path = []
        self.layout_hint = ''

    @property
    def position(self):
        box = self.box
        return wx.Point2D(box.x, box.y)
    @position.setter
    def position(self, point):
        #TODO: Move path based on box
        pass

    @property
    def box(self):
        gc = GraphicsContext.CreateMeasuringContext()
        graphics_path = self.make_path(gc)
        box = graphics_path.GetBox()
        if wx.Platform == '__WXGTK__':
            box = wx.Rect2D(*box)
            box.Inset(1, 1)
        enlarge = int((self.stroke_width-1.0)/2.0) # Convert to int to emulate where outline is drawn
        box.Inset(-enlarge, -enlarge)
        box.x = round(box.x) # Rounding necessary on MSW
        box.y = round(box.y)
        box.width = round(box.width)
        box.height = round(box.height)
        return box

    def make_path(self, gc):
        graphics_path = gc.CreatePath()
        for segment in self.path:
            segment.draw(graphics_path)
        return graphics_path

    def morph_h(self, x, ref_segment, segments):
        dx = x - ref_segment.point.x
        for segment in segments:
            segment.move_by(dx, 0)

    def morph_v(self, y, ref_segment, segments):
        dy = y - ref_segment.point.y
        for segment in segments:
            segment.move_by(0, dy)

    def draw(self, gc):
        gc.PushState()
        gc.SetPen(wx.Pen(self.stroke, self.stroke_width))
        gc.SetBrush(wx.Brush(self.fill))
        graphics_path = self.make_path(gc)
        gc.DrawPath(graphics_path)
        gc.PopState()

    def from_svg(self, svg_path):
        value = Word(nums + '.-+')
        value.setParseAction(lambda s,l,t : float(t[0]))

        point = value + Suppress(',') + value
        point.setParseAction(Point.from_svg)

        move = Word('Mm', max =1) + point
        move.setParseAction(MoveTo.from_svg)

        curve = Word('Cc', max=1) + point + point + point
        curve.setParseAction(CurveTo.from_svg)

        line = Word('Ll', max=1) + point
        line.setParseAction(LineTo.from_svg)

        close = Word('Zz', max=1)
        close.setParseAction(ClosePath.from_svg)

        tx_data = OneOrMore(move | curve | line | close)

        return tx_data.parseString(svg_path)


class Text(Shape):
    "Text using standard system font."
    def __init__(self, x, y, text='', fill=(0,0,0,255)):
        super(Text, self).__init__(x, y, 0, 0, fill=fill)
        if lang.LANGAUGE == 'chinese':
            size = 16
            if wx.Platform == '__WXMSW__':
                self.font = wx.Font(size, wx.DEFAULT, wx.NORMAL, wx.NORMAL, faceName='Microsoft YaHei')
            else:
                self.font = wx.SystemSettings.GetFont(wx.SYS_DEFAULT_GUI_FONT)
        else:
            size = 12
            self.font = wx.SystemSettings.GetFont(wx.SYS_DEFAULT_GUI_FONT)
            self.font.SetWeight(wx.BOLD)
        if wx.Platform == '__WXMSW__':
            self.font.SetPixelSize((0,-size)) # Negative number tells windows to use cell height c.f. GTK and Mac
        else:
            self.font.SetPointSize(float(size)) # Point = pixels on GTK and Mac
        self.text = text
        self.parent = None

    @property
    def box(self):
        gc = GraphicsContext.CreateMeasuringContext()
        gc.SetFont(self.font, self.fill)
        width, height, descent, externalLeading = gc.GetFullTextExtent(self.text)
        box = wx.Rect2D(self.x, self.y, width, height)
        box.x = round(box.x) # Rounding necessary on MSW
        box.y = round(box.y)
        box.width = round(box.width)
        box.height = round(box.height)
        return box

    def draw_shape(self, gc):
        gc.SetFont(self.font, self.fill)
        gc.DrawText(self.text, self.x, self.y)


class AttachPoint(wx.Point2D):
    def __init__(self, x, y, type_strings, parent):
        super(AttachPoint, self).__init__(x, y)
        self.type_strings = type_strings    # List of types of slave that will attach to
        self.attached_to = None
        self.parent = parent


class Block(Shape):
    "Block caches its children on a bitmap for faster rendering. Children can be blocks as well as other shapes."
    def __init__(self, x, y):
        self.children = []
        super(Block, self).__init__(x, y, 1, 1)
        self.draggable = False
        self.cache = None
        self.attach_offset = None
        self.type_string = ''
        self.attached_to = None
        self.attach_points = []
        self.adopt_points = []
        self.last_attach_point_reversable = False

    @property
    def canvas(self):
        return self._canvas

    @canvas.setter
    def canvas(self, value):
        self._canvas = value
        for child in self.children:
            child.canvas = value

    @property
    def box(self):
        return wx.Rect2D(self.x, self.y, self.width, self.height)

    def add(self, new_child):
        new_child.parent = self
        new_child.canvas = self.canvas
        self.children.append(new_child)

    def remove(self, child):
        child.parent = None
        child.canvas = None
        self.children.remove(child)

    def move(self, point):
        box = self.box
        self.position = wx.Point2DCopy(point)
        box.Union(self.box)
        return box

    def layout(self):
        pass

    def distribute_hozontally(self, children, top=0, left=0, top_border=0, bottom_border=0, left_border=0,
                              right_border=0, min_height=1, min_width=1):
        height = min_height
        width = min_width

        for child in children:
            if child.visible:
                child_height = child.box.height + top_border + bottom_border
                if child_height > height:
                    height = child_height

        x = left + left_border
        for child in children:
            if child.visible:
                box = child.box
                child.y = top + (height - box.height - top_border - bottom_border) / 2 + top_border
                child.x = x
                x += box.width

        if x + right_border - left > width:
            width = x + right_border - left

        return width, height

    def render(self, rect=None):
        old_box = self.box
        self.layout()
        box = wx.Rect2D()
        for child in self.children:
            try:
                if child.visible:
                    box = box.CreateUnion(child.box)
            except AttributeError:
                box = box.CreateUnion(child.box)
        box.SetLeftTop((0,0))
        self.width = box.width
        self.height = box.height
        rect = old_box.CreateUnion(self.box)
        self.width = box.width
        self.height = box.height
        if wx.Platform == '__WXGTK__':
            self.cache = GraphicsBitmap(int(self.width) + 1, int(self.height) + 1) # Add 1 for 0th pixel
            gc = GraphicsContext.CreateFromSurface(self.cache.Surface)
            gc.Clear((0,0,0,0))
            for child in self.children:
                child.draw(gc)
        else:
            self.cache = wx.EmptyBitmapRGBA(self.width + 1.0, self.height + 1.0, 0, 0, 0, 0) # Add 1 for 0th pixel
            dc = wx.MemoryDC(self.cache)
            gc = GraphicsContext.Create(dc)
            for child in self.children:
                child.draw(gc)
            dc.SelectObject(wx.NullBitmap)
        for attach_point in self.attach_points:
            if attach_point.attached_to:
                rect.Union(self.canvas.move_all_attached(attach_point.attached_to, self.position + attach_point - attach_point.attached_to.attach_offset))
        if self.parent:
            self.parent.render(rect)

    def draw(self, gc):
        if self.cache and self.visible:
            gc.DrawBitmap(self.cache, self.x, self.y, self.cache.Width, self.cache.Height)

    def contains(self, point, border=None):
        if self.visible:
            box = self.box
            if border:
                box.Inset(-border, -border)
            if box.Contains(point):
                if wx.Platform == '__WXGTK__':
                    stride = self.cache.Surface.get_stride()
                    data = self.cache.Surface.get_data()
                    x = int(point.x - box.x)
                    y = int(point.y - box.y)
                    byte_index = y*stride + x * 4
                    if ord(data[byte_index+3]) != 0:
                        return True
                    else:
                        return False
                else:
                    pixel_data = wx.AlphaPixelData(self.cache)
                    pixel_accessor = pixel_data.GetPixels()
                    pixel_accessor.MoveTo(pixel_data, point.x - box.x, point.y - box.y)
                    if pixel_accessor.Get()[3] != 0:
                        return True
                    else:
                        return False
            else:
                return False
        else:
            return False

    def on_click(self):
        pass

    def spawn(self):
        new_block = self.__class__(self.x, self.y)
        new_block.hit_offset = wx.Point2DCopy(self.hit_offset)
        return new_block


class Marker(Block):
    def __init__(self, x, y):
        super(Marker, self).__init__(x, y)
        self.add(Path("M24.5,6.5 L17.5,12.5 L17.5,9.5 L0.5,7.5 L0.5,5.5 L17.5,3.5 L17.5,0.5 L24.5,6.5 Z", fill=(0x00,0x00,0x00,0x7f), stroke=(0x00,0x00,0x00,0xFF)))
        self.visible = False
        self.render()

    def point_to(self, point):
        point -= wx.Point2D(self.width, self.height / 2)
        return self.move(point)

