#############################################################################
# Copyright (c) 2014-2017 Redfern Electronics Limited. All rights reserved. #
#############################################################################

from pyparsing import *

import assembler

ParserElement.setDefaultWhitespaceChars(' \t')

assembly = []
global_vars = []

# Expression
# ----------

def push_first(s, l, t):
    if t[0] in global_vars:
        # TODO: Error if variable doesn't exist
        var = t[0]
        address = global_vars.index(var)
        assembly.append('PUSH %d' % address)
    else:
        num = t[0]
        assembly.append('PUSHL %s' % num)

def push_op(s, l, t):
    op = t[0]
    if op == '+':
        assembly.append('ADD')
    elif op == '-':
        assembly.append('SUB')
    elif op == '*':
        assembly.append('MUL')
    elif op == '/':
        assembly.append('DIV')
    elif op == '<':
        assembly.append('LT')
    elif op == '>':
        assembly.append('GT')
    elif op == '==':
        assembly.append('EQ')
    elif op == 'and':
        assembly.append('AND')
    elif op == 'or':
        assembly.append('OR')

def push_not_op(s, l, t):
    if t and t[0] == 'not':
        assembly.append('NOT')

def push_uminus(s, l, t):
    if t and t[0] == '-':
        assembly.append('PUSHL -1')
        assembly.append('MUL')

num = Combine(Word('+-' + nums, nums))
ident = Word(alphas, alphas + nums + '_')
int_function = Forward()    # Forward declare functions that return int so they can be used in expressions.

lt = Literal('<')
gt = Literal('>')
eq = Literal('==')
and_op = Literal('and')
or_op = Literal('or')
not_op = Literal('not')
plus = Literal('+')
minus = Literal('-')
mul = Literal('*')
div = Literal('/')
lpar = Literal('(').suppress()
rpar = Literal(')').suppress()
comma = Literal(',').suppress()
logic_op = and_op | or_op
add_op = plus | minus
mul_op = mul | div
rel_op = lt | gt | eq

expression = Forward()

factor = int_function | (num|ident).setParseAction(push_first) | (lpar + expression.suppress() + rpar)
signed_factor = (Optional(add_op) + factor).setParseAction(push_uminus)
term = signed_factor + ZeroOrMore((mul_op + signed_factor).setParseAction(push_op))
num_expression = term + ZeroOrMore((add_op + term).setParseAction(push_op))
relation = num_expression + ZeroOrMore((rel_op + num_expression).setParseAction(push_op))
not_relation = (Optional(not_op) + relation).setParseAction(push_not_op)
expression << not_relation + ZeroOrMore((logic_op + not_relation).setParseAction(push_op))

# Variable declaration
# --------------------

def declare_var(s, l, t):
    var = t[1]
    # TODO: Check for redeclaration
    if var not in global_vars:
        global_vars.append(var)
        assembly.append('PUSHL 0')

declaration = Literal('int') + ident
declaration.setParseAction(declare_var)


# Assignment statement
# --------------------

def assign(s, l, t):
    # TODO: Check variable exists
    var = t[0]
    address = global_vars.index(var)
    assembly.append('POP %d' % address)

assignment = ident + '=' + expression
assignment.setParseAction(assign)


# Read digital input built-in function
# ------------------------------------

def compile_digital_read(s, l, t):
    assembly.append('DRD')

digital_read = (Literal('DigitalRead') + lpar + expression + rpar).setParseAction(compile_digital_read)


# Write digital output built-in function
# --------------------------------------

def compile_digital_read(s, l, t):
    assembly.append('DWR')

digital_write = (Literal('DigitalWrite') + lpar + expression + comma + expression + rpar).setParseAction(compile_digital_read)


# Read analogue input built-in function
# -------------------------------------

def compile_analogue_read(s, l, t):
    assembly.append('ARD')

analogue_read = (Literal('AnalogueRead') + lpar + expression + rpar).setParseAction(compile_analogue_read)


# Random integer function
# -----------------------

def compile_random(s, l, t):
    assembly.append('RND')

random = (Literal('Random') + lpar + expression + comma + expression + rpar).setParseAction(compile_random)


# Motor built-in functions
# ------------------------

def compile_motor1(s, l, t):
    assembly.append('MOT1')

def compile_motor2(s, l, t):
    assembly.append('MOT2')

motor1 = (Literal('SetMotor1') + lpar + expression + rpar).setParseAction(compile_motor1)
motor2 = (Literal('SetMotor2') + lpar + expression + rpar).setParseAction(compile_motor2)


# Sparkle built-in functions
# --------------------------

def compile_set_sparkle(s, l, t):
    assembly.append('SSPRK')
    assembly.append('OSPRK')

set_sparkle = (Literal('SetSparkle') + lpar + expression + comma + expression + comma + expression + comma + expression + rpar).setParseAction(compile_set_sparkle)


def compile_set_sparkle_buf(s, l, t):
    assembly.append('SSPRK')

set_sparkle_buf = (Literal('SetSparkleBuf') + lpar + expression + comma + expression + comma + expression + comma + expression + rpar).setParseAction(compile_set_sparkle_buf)


def compile_sparkle_out(s, l, t):
    assembly.append('OSPRK')

sparkle_out = (Literal('SparkleOut') + lpar + rpar).setParseAction(compile_sparkle_out)


# Servo built-in functions
# --------------------------

def compile_servo(s, l, t):
    assembly.append('SRV')

servo = (Literal('Servo') + lpar + expression + comma + expression + rpar).setParseAction(compile_servo)


# Ultrasonic distance sensor HC-SR04 built-in functions
# -----------------------------------------------------

def compile_distance(s, l, t):
    assembly.append('USONS')
    assembly.append('USONL')

distance = (Literal('Distance') + lpar + expression + comma + expression + rpar).setParseAction(compile_distance)


# Placeholder function
# --------------------

pass_func = Literal('pass')


# Forward declaration of statement so it can be used within control blocks
# ------------------------------------------------------------------------

statement = Forward()


# Label generator
# ---------------

label_count = 0
labels = []

def new_label():
    global label_count
    label = 'label_%d' % label_count
    label_count += 1
    labels.append(label)
    return label


# If
# --

def compile_if(s, l, t):
    label = new_label()
    assembly.append('BEZ %s' % label)

def compile_else(s, l, t):
    else_lable = labels.pop()
    end_label = new_label()
    assembly.append('BRA %s' % end_label)
    assembly.append(':%s' % else_lable)

def compile_if_end(s, l, t):
    assembly.append(':%s' % labels.pop())

if_block = ((Literal('if') + expression + Literal('then')).setParseAction(compile_if) + OneOrMore(statement) + Optional(Literal('else').setParseAction(compile_else) + OneOrMore(statement)) + Literal('end if')).setParseAction(compile_if_end)


# Do forever loop
# ---------------

def compile_forever_start(s, l, t):
    label = new_label()
    assembly.append(':%s' % label)

def compile_forever_end(s, l, t):
    assembly.append('BRA %s' % labels.pop())

forever_block = ((Literal('do') + Literal('forever') + LineEnd()).setParseAction(compile_forever_start) + OneOrMore(statement) + Literal('loop')).setParseAction(compile_forever_end)


# Do while loop
# -------------

def compile_while_start(s, l, t):
    start_label = new_label()
    assembly.append(':%s' % start_label)

def compile_while_test(s, l, t):
    exit_label = new_label()
    assembly.append('BEZ %s' % exit_label)

def compile_while_end(s, l, t):
    exit_label = labels.pop()
    start_label = labels.pop()
    assembly.append('BRA %s' % start_label)
    assembly.append(':%s' % exit_label)

while_block = ((Literal('do') + Literal('while').setParseAction(compile_while_start) + expression + LineEnd()).setParseAction(compile_while_test) + OneOrMore(statement) + Literal('loop')).setParseAction(compile_while_end)


# Do until loop
# -------------

def compile_until_start(s, l, t):
    start_label = new_label()
    assembly.append(':%s' % start_label)

def compile_until_test(s, l, t):
    exit_label = new_label()
    assembly.append('BNZ %s' % exit_label)

def compile_until_end(s, l, t):
    exit_label = labels.pop()
    start_label = labels.pop()
    assembly.append('BRA %s' % start_label)
    assembly.append(':%s' % exit_label)

until_block = ((Literal('do') + Literal('until').setParseAction(compile_until_start) + expression + LineEnd()).setParseAction(compile_until_test) + OneOrMore(statement) + Literal('loop')).setParseAction(compile_until_end)


# Do times loop
# -------------

def compile_repeat_start(s, l, t):
    start_label = new_label()
    exit_label = new_label()
    assembly.append(':%s' % start_label)
    assembly.append('DUP')
    assembly.append('BEZ %s' % exit_label)
    assembly.append('PUSHL 1')
    assembly.append('SUB')

def compile_repeat_end(s, l, t):
    exit_label = labels.pop()
    start_label = labels.pop()
    assembly.append('BRA %s' % start_label)
    assembly.append(':%s' % exit_label)
    assembly.append('POPRM')

repeat_block = ((Literal('do') + expression + Literal('times') + LineEnd()).setParseAction(compile_repeat_start) + OneOrMore(statement) + Literal('loop')).setParseAction(compile_repeat_end)


# Wait built-in function
# ----------------------

def compile_wait(s, l, t):
    # TODO: Doesn't deal with Wait(0)
    start_label = new_label()
    exit_label = new_label()
    assembly.append(':%s' % start_label)
    assembly.append('DUP')
    assembly.append('BEZ %s' % exit_label)
    assembly.append('WAIT')
    assembly.append('PUSHL 1')
    assembly.append('SUB')
    assembly.append('BRA %s' % start_label)
    assembly.append(':%s' % exit_label)
    assembly.append('POPRM')
    labels.pop()
    labels.pop()

wait = (Literal('Wait') + lpar + expression + rpar).setParseAction(compile_wait)


# Built-in functions by type
# --------------------------

int_function << (digital_read | digital_write | analogue_read | random | distance)
void_function = (wait | motor1 | motor2 | set_sparkle | set_sparkle_buf | sparkle_out | servo)


# Code
# ----

statement << (while_block | until_block | forever_block | repeat_block |  if_block |
              declaration | assignment |
              int_function | void_function |
              pass_func | Empty() + LineEnd())

code = ZeroOrMore(statement)


def compile(source):
    global assembly
    global global_vars
    assembly = []
    global_vars =[]
    results = code.parseString(source, parseAll=True)
    assembly.append('STOP')
    return assembler.assemble(assembly)