#############################################################################
# Copyright (c) 2014-2017 Redfern Electronics Limited. All rights reserved. #
#############################################################################

import platform
if platform.system() == 'Linux':
    import crumble_hidapi as hid
else:
    import hid

VERSION_CMD = 0
READ_CMD = 1
ERASE_CMD = 2
WRITE_CMD = 3
INIT_CMD = 4
SEND_VM_CMD_CMD = 5
READ_RAM_CMD = 6
WRITE_RAM_CMD = 7

VM_POLL = 0
VM_START = 1
VM_STOP = 2
VM_STEP = 3

VM_START_ADDRESS = 0x1000
BYTE_CODE_START = 0x1C00
BYTE_CODE_LENGTH = 1024

VENDOR_ID = 0x1C40
PRODUCT_ID = 0x05B8

device = hid.device()
connected = False

def connect():
    """Attempt to connect to HID."""
    global connected
    try:
        device.open(VENDOR_ID, PRODUCT_ID)
    except IOError:
        connected = False
        raise
    else:
        connected = True

def read(length):
    """Read data from HID.
    Returns data as list of bytes."""
    global connected
    if not connected:
        connect()
    if connected:
        try:    # Can optimize to just catch exception when only using hidapi via ctypes
            data = device.read(length)
        except IOError:
            data = []
        if not data:
            connected = False
            return []
        else:
            return data[0:length]
    else:
        return []

def write(data):
    """Writes data to HID.
    data -- list of bytes."""
    global connected
    if not connected:
        connect()
    if connected:
        try:    # Can optimize to just catch exception when only using hidapi via ctypes
            result = device.write([0] + data)
        except IOError:
            result = -1
        if result==-1:
            connected = False
        return result
    else:
        return -1

def read_flash(address, length):
    """Read flash data.
    address -- Address of first word.
    length -- Number of words to read
    Returns data as list of words."""
    tx_data = [READ_CMD,
               address & 0xFF,
               (address>>8) & 0xFF,
               length]
    write(tx_data)
    rx_data = read(64)
    data = []
    for i in range(0,len(rx_data),2):
        data.append(rx_data[i] + (rx_data[i+1]<<8))
    return data[:length]

def erase_flash_row(address):
    """Erase flash row (32 words) containing address."""
    tx_data = [ERASE_CMD,
               address & 0xFF,
               (address>>8) & 0xFF]
    write(tx_data)
    rx_data = read(64)
#         print 'Erase success:', rx_data[0], address & 0xFF, (address>>8) & 0xFF

def write_flash(address, data):
    """Write flash data.
    address -- word address.
    data -- list of words."""
    tx_data = [WRITE_CMD,
               address & 0xFF,
               (address>>8) & 0xFF,
               len(data)]

    for w in data:
        tx_data += [w & 0xFF, (w >> 8) & 0xFF]

    write(tx_data)
    rx_data = read(64)
#         print 'Write success:', rx_data[0]

def program_hex_from_file(filename):
    """Reads hex file formatted with 16 word per line and
    programs flash."""
    hex_file = open(filename)
    current_row = -1

    for line in hex_file.readlines():
        if line[:3]==':10' and line[7:9]=='00':
            address = int(line[3:7],16) / 2
            if address & ~0x001F != current_row:
                current_row = address & ~0x001F
                erase_flash_row(current_row)
            data = []
            for i in range(9, 41, 4):
                data.append(int(line[i+2:i+4] + line[i:i+2], 16))
            write_flash(address, data)

    hex_file.close()

def program_hex_old(hex_data):
    """Reads form list of hex strings formatted with 16 word per line and
    programs flash."""
    current_row = -1
    for line in hex_data:
        if line[:3]==':10' and line[7:9]=='00':
            address = int(line[3:7],16) / 2
            if address & ~0x001F != current_row:
                current_row = address & ~0x001F
                erase_flash_row(current_row)
            data = []
            for i in range(9, 41, 4):
                data.append(int(line[i+2:i+4] + line[i:i+2], 16))
            write_flash(address, data)

def erase_vm_first_page():
    erase_flash_row(VM_START_ADDRESS)

def program_hex(hex_data, first_page_only=False):
    """Reads from list of hex strings and programs flash."""
    page_size = 32  # Page size words
    data = {}
    for line in hex_data:
        num_bytes = int(line[1:3], 16)
        address = int(line[3:7], 16) / 2
        type = int(line[7:9], 16)

        if num_bytes > 0 and type == 0:
            for i in range(9, 9 + num_bytes * 2, 4):
                page_address = address & ~(page_size - 1)
                if page_address not in data:
                    data[page_address] = [0x3FFF] * page_size
                data[page_address][address - page_address] = int(line[i+2:i+4] + line[i:i+2], 16)
                address += 1

    if first_page_only:
        for page_address in data.keys():
            if page_address != VM_START_ADDRESS:
                del(data[page_address])

    # Erase all memory first (to avoid leaving broken VM)
    for page_address in sorted(data.keys()):
        erase_flash_row(page_address)

    # Program hex in reverse (VM version then gets programmed last, so Crumble wont start if incomplete)
    for page_address in sorted(data.keys(), reverse=True):
        for address in sorted(range(page_address, page_address + page_size, 16), reverse=True):
            write_flash(address, data[page_address][address - page_address:address - page_address + 16])

def init_vm():
    """Resets the VM state."""
    write([INIT_CMD, VM_STOP])
    rx_data = read(64)
    # print 'VM initialisation success:', rx_data[0]

def stop_vm():
    """Stop the VM running."""
    send_vm_command(VM_STOP)

def send_vm_command(command):
    """Sends a command directly to the VM."""
    write([SEND_VM_CMD_CMD, command])
    rx_data = read(64)
    # print 'Command send success:', rx_data[0]

def read_ram(address, length):
    """Read vm ram.
    address -- Address of first byte.
    length -- Number of bytes to read
    Returns data as list of bytes."""
    tx_data = [READ_RAM_CMD,
               address & 0xFF,
               (address>>8) & 0xFF,
               length]

    write(tx_data)
    rx_data = read(length)
    return rx_data

def write_ram(address, data):
    """Write to vm ram.
    address -- byte address.
    data -- list of bytes."""
    tx_data = [WRITE_RAM_CMD,
               address & 0xFF,
               (address>>8) & 0xFF,
               len(data)] + data
    write(tx_data)
    rx_data = read(64)

def read_vm_version():
    """Read VM version from flash."""
    data = read_flash(0x1000,1)
    if data:
        return data[0] & 0xFF
    else:
        return -1

def write_byte_code(byte_code):
    """Write byte code to flash."""
    if len(byte_code) > BYTE_CODE_LENGTH:
        raise IOError('Trying to write byte code that is too large.')

    width = 16
    r = len(byte_code) % width
    if r:
        byte_code += [0x3FFF] * (width - r)

    current_row = -1
    address = BYTE_CODE_START

    for i in range(0, len(byte_code), width):
        if address & ~0x001F != current_row:
            current_row = address & ~0x001F
            erase_flash_row(current_row)
        write_flash(address, byte_code[i:i + width])
        address += width