# -*- coding: utf-8 -*-

#############################################################################
# Copyright (c) 2014-2017 Redfern Electronics Limited. All rights reserved. #
#############################################################################

from lang import _
import blocks
import canvas
from weakref import WeakSet
import wx
try:
    from pubsub import pub
except ImportError:
    if wx.VERSION[0] < 3:
        from wx.lib.pubsub import setupkwargs
    from wx.lib.pubsub import pub


pins = {0:0, 1:1, 2:3, 3:2}

class WhenRunBlock(blocks.WhenBlock, blocks.XMLBlock):
    def __init__(self, x, y):
        super(WhenRunBlock, self).__init__(x, y, _('program start'))

class DigitalInputBlock(blocks.BooleanBlock, blocks.XMLBlock):
    def __init__(self, x, y):
        super(DigitalInputBlock, self).__init__(x, y)
        text = _('|is|').split('|')
        self.outline.fill = blocks.io_colour
        self.abcd_entry = blocks.ToggledEntryBlock(0, 0, ['A', 'B', 'C', 'D'])
        self.add(self.abcd_entry)
        self.add(canvas.Space(0, 0 , 6, 5))
        self.add(canvas.Text(0, 0, text[1], (255,255,255,255)))
        self.add(canvas.Space(0, 0 , 6, 5))
        self.hi_lo_entry = blocks.ToggledEntryBlock(0, 0)
        self.add(self.hi_lo_entry)
        self.render()

    def to_xml(self, name=None, position=False):
        element = blocks.XMLBlock.to_xml(self, name, position)
        element.append(self.abcd_entry.to_xml('pin'))
        element.append(self.hi_lo_entry.to_xml('state'))
        return element

    def from_xml(self, element, canvas_dest):
        blocks.XMLBlock.from_xml(self, element, canvas_dest)
        self.abcd_entry.from_xml(element.find('pin'), canvas_dest)
        self.hi_lo_entry.from_xml(element.find('state'), canvas_dest)

    def basic_code(self, inits, setups):
        basic_code = '('
        if self.hi_lo_entry.option == 'LO':
            basic_code += 'not '
        basic_code += 'DigitalRead(%d))' % pins[self.abcd_entry.index]
        return basic_code


class AnalogueInput(blocks.ValueBlock, blocks.XMLBlock):
    def __init__(self, x, y):
        super(AnalogueInput, self).__init__(x, y)
        text = _('analogue|').split('|')
        self.outline.fill = blocks.io_colour
        self.add(canvas.Text(0, 0, text[0], (255,255,255,255)))
        self.add(canvas.Space(0, 0 , 6, 5))
        self.abcd_entry = blocks.ToggledEntryBlock(0, 0, ['A', 'B', 'C', 'D'])
        self.add(self.abcd_entry)
        self.render()

    def to_xml(self, name=None, position=False):
        element = blocks.XMLBlock.to_xml(self, name, position)
        element.append(self.abcd_entry.to_xml('pin'))
        return element

    def from_xml(self, element, canvas_dest):
        blocks.XMLBlock.from_xml(self, element, canvas_dest)
        self.abcd_entry.from_xml(element.find('pin'), canvas_dest)

    def basic_code(self, inits, setups):
        basic_code = ''
        basic_code += '(AnalogueRead(%d) * 0 + AnalogueRead(%d)/4)' % (pins[self.abcd_entry.index], pins[self.abcd_entry.index]) # TODO: Remove extra read once firmware fixed
        return basic_code

class DigitalOutputBlock(blocks.IOStatementBlock, blocks.XMLBlock):
    def __init__(self, x, y):
        super(DigitalOutputBlock, self).__init__(x, y)
        text = _('set||').split('|')
        self.add(canvas.Text(0, 0, text[0], (255,255,255,255)))
        self.add(canvas.Space(0, 0 , 6, 5))
        self.abcd_entry = blocks.ToggledEntryBlock(0, 0, ['A', 'B', 'C', 'D'])
        self.add(self.abcd_entry)
        self.add(canvas.Space(0, 0 , 6, 5))
        self.hi_lo_entry = blocks.ToggledEntryBlock(0, 0)
        self.add(self.hi_lo_entry)
        self.render()

    def to_xml(self, name=None, position=False):
        element = blocks.XMLBlock.to_xml(self, name, position)
        element.append(self.abcd_entry.to_xml('pin'))
        element.append(self.hi_lo_entry.to_xml('state'))
        return element

    def from_xml(self, element, canvas_dest):
        blocks.XMLBlock.from_xml(self, element, canvas_dest)
        self.abcd_entry.from_xml(element.find('pin'), canvas_dest)
        self.hi_lo_entry.from_xml(element.find('state'), canvas_dest)

    def basic_code(self, inits, setups):
        return 'DigitalWrite(%d, %d)\n' % (pins[self.abcd_entry.index], 1 - self.hi_lo_entry.index)


class MotorBlock(blocks.IOStatementBlock, blocks.XMLBlock):
    def __init__(self, x, y):
        super(MotorBlock, self).__init__(x, y)
        text = _('motor|at|%').split('|')
        self.add(canvas.Text(0, 0, text[0], (255,255,255,255)))
        self.add(canvas.Space(0, 0 , 6, 5))
        self.motor = blocks.ToggledEntryBlock(0, 0, ['1', '2'])
        self.add(self.motor)
        self.add(canvas.Space(0, 0 , 6, 5))
        self.action = blocks.ToggledEntryBlock(0, 0, ['FORWARD', 'REVERSE', 'STOP'])
        self.action.index = 0
        self.add(self.action)

        self.add(canvas.Space(0, 0 , 6, 5))
        self.add(canvas.Text(0, 0, text[1], (255,255,255,255)))
        self.add(canvas.Space(0, 0 , 6, 5))
        self.speed = blocks.ValueEntryBlock(0, 0, 75, min_value=-100, max_value=100)
        self.add(self.speed)
        self.add(canvas.Space(0, 0 , 6, 5))
        self.add(canvas.Text(0, 0, text[2], (255,255,255,255)))

        self.hideable = self.children[-6:]

        self.render()

    def render(self, rect=None):
        if self.action.option == 'STOP':
            for child in self.hideable:
                child.visible = False
        else:
            for child in self.hideable:
                child.visible = True
        super(MotorBlock, self).render(rect)

    def to_xml(self, name=None, position=False):
        element = blocks.XMLBlock.to_xml(self, name, position)
        element.append(self.motor.to_xml('motor'))
        element.append(self.action.to_xml('action'))
        element.append(self.speed.to_xml('speed'))
        return element

    def from_xml(self, element, canvas_dest):
        blocks.XMLBlock.from_xml(self, element, canvas_dest)
        self.motor.from_xml(element.find('motor'), canvas_dest)
        self.action.from_xml(element.find('action'), canvas_dest)
        self.speed.from_xml(element.find('speed'), canvas_dest)

    def basic_code(self, inits, setups):
        if self.action.option == 'FORWARD':
            return 'SetMotor%s((%s * 187) / 100)\n' % (self.motor.option, self.speed.basic_code(inits, setups))
        elif self.action.option == 'REVERSE':
            return 'SetMotor%s((-%s * 187) / 100)\n' % (self.motor.option, self.speed.basic_code(inits, setups))
        else:
            return 'SetMotor%s(0)\n' % self.motor.option


def MotorSpeedBlock(x, y):
    """For compatibility with 0.13.7 and earlier"""
    motor_block = MotorBlock(x, y)
    motor_block.action.option = 'FORWARD'
    return motor_block


class ServoBlock(blocks.IOStatementBlock, blocks.XMLBlock):
    def __init__(self, x, y):
        super(ServoBlock, self).__init__(x, y)
        text = _('servo|degrees').split('|')
        self.add(canvas.Text(0, 0, text[0], (255,255,255,255)))
        self.add(canvas.Space(0, 0 , 6, 5))
        self.abcd_entry = blocks.ToggledEntryBlock(0, 0, ['A', 'B', 'C', 'D'])
        self.add(self.abcd_entry)
        self.add(canvas.Space(0, 0 , 6, 5))
        self.angle_entry = blocks.ValueEntryBlock(0, 0, 0, min_value=-90, max_value=90)
        self.add(self.angle_entry)
        self.add(canvas.Space(0, 0 , 6, 5))
        self.add(canvas.Text(0, 0, text[1], (255,255,255,255)))
        self.render()

    def to_xml(self, name=None, position=False):
        element = blocks.XMLBlock.to_xml(self, name, position)
        element.append(self.abcd_entry.to_xml('pin'))
        element.append(self.angle_entry.to_xml('angle'))
        return element

    def from_xml(self, element, canvas_dest):
        blocks.XMLBlock.from_xml(self, element, canvas_dest)
        self.abcd_entry.from_xml(element.find('pin'), canvas_dest)
        self.angle_entry.from_xml(element.find('angle'), canvas_dest)

    def basic_code(self, inits, setups):
        counts_per_deg = 34
        centre = 4545
        return 'Servo(%d, %d + %s * %d)\n' % (pins[self.abcd_entry.index], centre, self.angle_entry.basic_code(inits, setups), counts_per_deg)


class DistanceBlock(blocks.ValueBlock, blocks.XMLBlock):
    def __init__(self, x, y):
        super(DistanceBlock, self).__init__(x, y)
        text = _('distance (cm)|').split('|')
        self.outline.fill = blocks.io_colour
        self.add(canvas.Text(0, 0, text[0], (255,255,255,255)))
        self.add(canvas.Space(0, 0 , 12, 5))
        self.add(canvas.Text(0, 0, 'T:', (255,255,255,255)))
        self.trig_abcd_entry = blocks.ToggledEntryBlock(0, 0, ['A', 'B', 'C', 'D'])
        self.add(self.trig_abcd_entry)
        self.add(canvas.Space(0, 0 , 6, 5))
        self.add(canvas.Text(0, 0, 'E:', (255,255,255,255)))
        self.echo_abcd_entry = blocks.ToggledEntryBlock(0, 0, ['A', 'B', 'C', 'D'])
        self.echo_abcd_entry.on_click()
        self.add(self.echo_abcd_entry)
        self.render()

    def to_xml(self, name=None, position=False):
        element = blocks.XMLBlock.to_xml(self, name, position)
        element.append(self.trig_abcd_entry.to_xml('trig'))
        element.append(self.echo_abcd_entry.to_xml('echo'))
        return element

    def from_xml(self, element, canvas_dest):
        blocks.XMLBlock.from_xml(self, element, canvas_dest)
        self.trig_abcd_entry.from_xml(element.find('trig'), canvas_dest)
        self.echo_abcd_entry.from_xml(element.find('echo'), canvas_dest)

    def basic_code(self, inits, setups):
        basic_code = ''
        basic_code += '(Distance(%d, %d))' % (pins[self.trig_abcd_entry.index], pins[self.echo_abcd_entry.index])
        return basic_code


class SetSprakleBlock(blocks.SparkleStatementBlcok, blocks.XMLBlock):
    def __init__(self, x, y):
        super(SetSprakleBlock, self).__init__(x, y)
        text =_('set sparkle|to|').split('|')
        self.add(canvas.Text(0, 0, text[0], (0xFF,0xFF,0xFF,0xFF)))
        self.add(canvas.Space(0, 0 , 6, 5))
        self.sparkle_index = blocks.ValueEntryBlock(0, 0, 0, min_value=0, max_value=31)
        self.add(self.sparkle_index)
        self.add(canvas.Space(0, 0 , 6, 5))
        self.add(canvas.Text(0, 0, text[1], (0xFF,0xFF,0xFF,0xFF)))
        self.add(canvas.Space(0, 0 , 6, 5))
        self.colour_entry = blocks.ColourEntryBlock(0, 0)
        self.add(self.colour_entry)
        self.render()

    def to_xml(self, name=None, position=False):
        element = blocks.XMLBlock.to_xml(self, name, position)
        element.append(self.sparkle_index.to_xml('address'))
        element.append(self.colour_entry.to_xml('colour'))
        return element

    def from_xml(self, element, canvas_dest):
        blocks.XMLBlock.from_xml(self, element, canvas_dest)
        self.sparkle_index.from_xml(element.find('address'), canvas_dest)
        self.colour_entry.from_xml(element.find('colour'), canvas_dest)

    def basic_code(self, inits, setups):
        return 'SetSparkle(%s, %d, %d, %d)\n' % (self.sparkle_index.basic_code(inits, setups),
                                                 self.colour_entry.colour[0],
                                                 self.colour_entry.colour[1],
                                                 self.colour_entry.colour[2])


class TurnSprakleOffBlock(blocks.SparkleStatementBlcok, blocks.XMLBlock):
    def __init__(self, x, y):
        super(TurnSprakleOffBlock, self).__init__(x, y)
        text = _('turn sparkle|off').split('|')
        self.add(canvas.Text(0, 0, text[0], (0xFF,0xFF,0xFF,0xFF)))
        self.add(canvas.Space(0, 0 , 6, 5))
        self.sparkle_index = blocks.ValueEntryBlock(0, 0, 0, min_value=0, max_value=31)
        self.add(self.sparkle_index)
        self.add(canvas.Space(0, 0 , 6, 5))
        self.add(canvas.Text(0, 0, text[1], (0xFF,0xFF,0xFF,0xFF)))
        self.render()

    def to_xml(self, name=None, position=False):
        element = blocks.XMLBlock.to_xml(self, name, position)
        element.append(self.sparkle_index.to_xml('address'))
        return element

    def from_xml(self, element, canvas_dest):
        blocks.XMLBlock.from_xml(self, element, canvas_dest)
        self.sparkle_index.from_xml(element.find('address'), canvas_dest)

    def basic_code(self, inits, setups):
        return 'SetSparkle(%s, 0, 0, 0)\n' % self.sparkle_index.basic_code(inits, setups)


class SetSprakleRGBBlock(blocks.SparkleStatementBlcok, blocks.XMLBlock):
    def __init__(self, x, y):
        super(SetSprakleRGBBlock, self).__init__(x, y)
        text = _('set sparkle|to|').split('|')
        self.add(canvas.Text(0, 0, text[0], (0xFF,0xFF,0xFF,0xFF)))
        self.add(canvas.Space(0, 0 , 6, 5))
        self.sparkle_index = blocks.ValueEntryBlock(0, 0, 0, min_value=0, max_value=31)
        self.add(self.sparkle_index)
        self.add(canvas.Space(0, 0 , 6, 5))
        self.add(canvas.Text(0, 0, text[1], (0xFF,0xFF,0xFF,0xFF)))
        self.add(canvas.Space(0, 0 , 6, 5))
        self.red = blocks.ValueEntryBlock(0, 0, 0, min_value=0, max_value=255)
        self.red.outline.fill = (255,192,192,255)
        self.red.render()
        self.add(self.red)
        self.add(canvas.Space(0, 0 , 6, 5))
        self.green = blocks.ValueEntryBlock(0, 0, 0, min_value=0, max_value=255)
        self.green.outline.fill = (192,255,192,255)
        self.green.render()
        self.add(self.green)
        self.add(canvas.Space(0, 0 , 6, 5))
        self.blue = blocks.ValueEntryBlock(0, 0, 0, min_value=0, max_value=255)
        self.blue.outline.fill = (192,192,255,255)
        self.blue.render()
        self.add(self.blue)
        self.render()

    def to_xml(self, name=None, position=False):
        element = blocks.XMLBlock.to_xml(self, name, position)
        element.append(self.sparkle_index.to_xml('address'))
        element.append(self.red.to_xml('red'))
        element.append(self.green.to_xml('green'))
        element.append(self.blue.to_xml('blue'))
        return element

    def from_xml(self, element, canvas_dest):
        blocks.XMLBlock.from_xml(self, element, canvas_dest)
        self.sparkle_index.from_xml(element.find('address'), canvas_dest)
        self.red.from_xml(element.find('red'), canvas_dest)
        self.green.from_xml(element.find('green'), canvas_dest)
        self.blue.from_xml(element.find('blue'), canvas_dest)

    def basic_code(self, inits, setups):
        return 'SetSparkle(%s, %s, %s, %s)\n' % (self.sparkle_index.basic_code(inits, setups),
                                                 self.red.basic_code(inits, setups),
                                                 self.green.basic_code(inits, setups),
                                                 self.blue.basic_code(inits, setups))


class SetAllSpraklesBlock(blocks.SparkleStatementBlcok, blocks.XMLBlock):
    def __init__(self, x, y):
        super(SetAllSpraklesBlock, self).__init__(x, y)
        text = _('set all sparkles to|').split('|')
        self.add(canvas.Text(0, 0, text[0], (0xFF,0xFF,0xFF,0xFF)))
        self.add(canvas.Space(0, 0 , 6, 5))
        self.colour_entry = blocks.ColourEntryBlock(0, 0)
        self.add(self.colour_entry)
        self.render()

    def to_xml(self, name=None, position=False):
        element = blocks.XMLBlock.to_xml(self, name, position)
        element.append(self.colour_entry.to_xml('colour'))
        return element

    def from_xml(self, element, canvas_dest):
        blocks.XMLBlock.from_xml(self, element, canvas_dest)
        self.colour_entry.from_xml(element.find('colour'), canvas_dest)

    def basic_code(self, inits, setups):
        inits['int all_sparkles_i\n'] = 1
        inits['int all_sparkles_r\n'] = 1
        inits['int all_sparkles_g\n'] = 1
        inits['int all_sparkles_b\n'] = 1

        basic_code = ''
        basic_code += 'all_sparkles_i = 0\n'
        basic_code += 'all_sparkles_r = %s\n' % self.colour_entry.colour[0]
        basic_code += 'all_sparkles_g = %s\n' % self.colour_entry.colour[1]
        basic_code += 'all_sparkles_b = %s\n' % self.colour_entry.colour[2]
        basic_code += 'do while all_sparkles_i < 32\n'
        basic_code += 'SetSparkleBuf(all_sparkles_i, all_sparkles_r, all_sparkles_g, all_sparkles_b)\n'
        basic_code += 'all_sparkles_i = all_sparkles_i + 1\n'
        basic_code += 'loop\n'
        basic_code += 'SparkleOut()\n'
        return basic_code


class TurnAllSpraklesOffBlock(blocks.SparkleStatementBlcok, blocks.XMLBlock):
    def __init__(self, x, y):
        super(TurnAllSpraklesOffBlock, self).__init__(x, y)
        text = _('turn all sparkles off').split('|')
        self.add(canvas.Text(0, 0, text[0], (0xFF,0xFF,0xFF,0xFF)))
        self.render()

    def to_xml(self, name=None, position=False):
        element = blocks.XMLBlock.to_xml(self, name, position)
        return element

    def from_xml(self, element, canvas_dest):
        blocks.XMLBlock.from_xml(self, element, canvas_dest)

    def basic_code(self, inits, setups):
        inits['int all_sparkles_i\n'] = 1
        basic_code = ''
        basic_code += 'all_sparkles_i = 0\n'
        basic_code += 'do while all_sparkles_i < 32\n'
        basic_code += 'SetSparkleBuf(all_sparkles_i, 0, 0, 0)\n'
        basic_code += 'all_sparkles_i = all_sparkles_i + 1\n'
        basic_code += 'loop\n'
        basic_code += 'SparkleOut()\n'
        return basic_code

class SetAllSparklesRGBBlock(blocks.SparkleStatementBlcok, blocks.XMLBlock):
    def __init__(self, x, y):
        super(SetAllSparklesRGBBlock, self).__init__(x, y)
        text = _('set all sparkles to|').split('|')
        self.add(canvas.Text(0, 0, text[0], (0xFF,0xFF,0xFF,0xFF)))
        self.add(canvas.Space(0, 0 , 6, 5))
        self.red = blocks.ValueEntryBlock(0, 0, 0, min_value=0, max_value=255)
        self.red.outline.fill = (255,192,192,255)
        self.red.render()
        self.add(self.red)
        self.add(canvas.Space(0, 0 , 6, 5))
        self.green = blocks.ValueEntryBlock(0, 0, 0, min_value=0, max_value=255)
        self.green.outline.fill = (192,255,192,255)
        self.green.render()
        self.add(self.green)
        self.add(canvas.Space(0, 0 , 6, 5))
        self.blue = blocks.ValueEntryBlock(0, 0, 0, min_value=0, max_value=255)
        self.blue.outline.fill = (192,192,255,255)
        self.blue.render()
        self.add(self.blue)
        self.render()

    def to_xml(self, name=None, position=False):
        element = blocks.XMLBlock.to_xml(self, name, position)
        element.append(self.red.to_xml('red'))
        element.append(self.green.to_xml('green'))
        element.append(self.blue.to_xml('blue'))
        return element

    def from_xml(self, element, canvas_dest):
        blocks.XMLBlock.from_xml(self, element, canvas_dest)
        self.red.from_xml(element.find('red'), canvas_dest)
        self.green.from_xml(element.find('green'), canvas_dest)
        self.blue.from_xml(element.find('blue'), canvas_dest)

    def basic_code(self, inits, setups):
        inits['int all_sparkles_i\n'] = 1
        inits['int all_sparkles_r\n'] = 1
        inits['int all_sparkles_g\n'] = 1
        inits['int all_sparkles_b\n'] = 1

        basic_code = ''
        basic_code += 'all_sparkles_i = 0\n'
        basic_code += 'all_sparkles_r = %s\n' % self.red.basic_code(inits, setups)
        basic_code += 'all_sparkles_g = %s\n' % self.green.basic_code(inits, setups)
        basic_code += 'all_sparkles_b = %s\n' % self.blue.basic_code(inits, setups)
        basic_code += 'do while all_sparkles_i < 32\n'
        basic_code += 'SetSparkleBuf(all_sparkles_i, all_sparkles_r, all_sparkles_g, all_sparkles_b)\n'
        basic_code += 'all_sparkles_i = all_sparkles_i + 1\n'
        basic_code += 'loop\n'
        basic_code += 'SparkleOut()\n'
        return basic_code


class WaitBlock(blocks.ControlStatementBlock, blocks.XMLBlock):
    def __init__(self, x, y):
        super(WaitBlock, self).__init__(x, y)
        text = _('wait|seconds').split('|')
        self.add(canvas.Text(0, 0, text[0], (0xFF,0xFF,0xFF,0xFF)))
        self.add(canvas.Space(0, 0 , 6, 5))
        self.time = blocks.ValueEntryBlock(0, 0, 1, min_value=0, precision=2)
        self.add(self.time)
        self.add(canvas.Space(0, 0 , 6, 5))
        self.add(canvas.Text(0, 0, text[1], (0xFF,0xFF,0xFF,0xFF)))
        self.render()

    def to_xml(self, name=None, position=False):
        element = blocks.XMLBlock.to_xml(self, name, position)
        element.append(self.time.to_xml('time'))
        return element

    def from_xml(self, element, canvas_dest):
        blocks.XMLBlock.from_xml(self, element, canvas_dest)
        self.time.from_xml(element.find('time'), canvas_dest)

    def basic_code(self, inits, setups):
        arg = self.time.basic_code(inits, setups)
        try:
            s = float(arg)
        except ValueError:
            basic_code = 'do %s times\n' % str(arg)
            basic_code += 'Wait(1000)\n'
            basic_code += 'loop\n'
        else:
            whole_s = int(s)
            ms = int(1000 * (s - whole_s))
            basic_code = ''
            if whole_s > 0:
                basic_code = 'do %s times\n' % str(whole_s)
                basic_code += 'Wait(1000)\n'
                basic_code += 'loop\n'
            if ms > 0:
                basic_code += 'Wait(%s)\n' % str(ms)
        return basic_code


class WaitMSBlock(blocks.ControlStatementBlock, blocks.XMLBlock):
    def __init__(self, x, y):
        super(WaitMSBlock, self).__init__(x, y)
        text = _('wait|milliseconds').split('|')
        self.add(canvas.Text(0, 0, text[0], (0xFF,0xFF,0xFF,0xFF)))
        self.add(canvas.Space(0, 0 , 6, 5))
        self.time = blocks.ValueEntryBlock(0, 0, 100, min_value=1)
        self.add(self.time)
        self.add(canvas.Space(0, 0 , 6, 5))
        self.add(canvas.Text(0, 0, text[1], (0xFF,0xFF,0xFF,0xFF)))
        self.render()

    def to_xml(self, name=None, position=False):
        element = blocks.XMLBlock.to_xml(self, name, position)
        element.append(self.time.to_xml('time'))
        return element

    def from_xml(self, element, canvas_dest):
        blocks.XMLBlock.from_xml(self, element, canvas_dest)
        self.time.from_xml(element.find('time'), canvas_dest)

    def basic_code(self, inits, setups):
        return 'Wait(%s)\n' % self.time.basic_code(inits, setups)


class WaitUntilBlock(blocks.ControlStatementBlock, blocks.XMLBlock):
    def __init__(self, x, y):
        super(WaitUntilBlock, self).__init__(x, y)
        text = _('wait until|').split('|')
        self.add(canvas.Text(0, 0, text[0], (0xFF,0xFF,0xFF,0xFF)))
        self.add(canvas.Space(0, 0 , 6, 5))
        self.conditional = blocks.BooleanEntryBlock(0, 0)
        self.add(self.conditional)
        self.render()

    def to_xml(self, name=None, position=False):
        element = blocks.XMLBlock.to_xml(self, name, position)
        element.append(self.conditional.to_xml('conditional'))
        return element

    def from_xml(self, element, canvas_dest):
        blocks.XMLBlock.from_xml(self, element, canvas_dest)
        self.conditional.from_xml(element.find('conditional'), canvas_dest)

    def basic_code(self, inits, setups):
        if self.conditional.adopt_point.attached_to:
            condition = self.conditional.adopt_point.attached_to.basic_code(inits, setups)
            basic_code = 'do until %s\n' % condition
            basic_code += 'pass\n'
            basic_code += 'loop\n'
            return basic_code
        else:
            return ''


class IfBlock(blocks.ControlBlock, blocks.XMLBlock):
    def __init__(self, x, y):
        super(IfBlock, self).__init__(x, y)
        text = _('if|then').split('|')
        text_bottom = _('end if').split('|')
        self.if_text = canvas.Text(0, 0, text[0], (0xFF,0xFF,0xFF,0xFF))
        self.if_text.layout_hint = 'top'
        self.add(self.if_text)
        self.if_space = canvas.Space(0, 0, 6, 6)
        self.if_space.layout_hint = 'top'
        self.add(self.if_space)
        self.conditional = blocks.BooleanEntryBlock(0,0)
        self.conditional.layout_hint = 'top'
        self.conditional_space = canvas.Space(0, 0, 6, 6)
        self.conditional_space.layout_hint = 'top'
        self.add(self.conditional)
        self.add(self.conditional_space)
        self.then_text = canvas.Text(0, 0, text[1], (0xFF,0xFF,0xFF,0xFF))
        self.then_text.layout_hint = 'top'
        self.add(self.then_text)
        self.end_text=canvas.Text(0, 0, text_bottom[0], (0xFF,0xFF,0xFF,0xFF))
        self.end_text.layout_hint = 'bottom'
        self.add(self.end_text)
        self.render()

    def to_xml(self, name=None, position=False):
        element = blocks.XMLBlock.to_xml(self, name, position)
        element.append(self.conditional.to_xml('conditional'))
        return element

    def from_xml(self, element, canvas_dest):
        blocks.XMLBlock.from_xml(self, element, canvas_dest)
        self.conditional.from_xml(element.find('conditional'), canvas_dest)

    def basic_code(self, inits, setups):
        if self.conditional.adopt_point.attached_to:
            condition = self.conditional.adopt_point.attached_to.basic_code(inits, setups)
        else:
            return ''
        basic_code = 'if %s then\n' % condition
        if self.adopt_point.attached_to:
            blocks = self.canvas.find_all_attached(self.adopt_point.attached_to)
            for block in blocks:
                basic_code += block.basic_code(inits, setups)
            basic_code += 'end if\n'
            return basic_code
        else:
            return ''


class IfElseBlock(blocks.ControlDoubleBlock, blocks.XMLBlock):
    def __init__(self, x, y):
        super(IfElseBlock, self).__init__(x, y)
        text = _('if|then').split('|')
        text_middle = _('else').split('|')
        text_bottom = _('end if').split('|')
        self.if_text = canvas.Text(0, 0, text[0], (0xFF,0xFF,0xFF,0xFF))
        self.if_text.layout_hint = 'top'
        self.add(self.if_text)
        self.if_space = canvas.Space(0, 0, 6, 6)
        self.if_space.layout_hint = 'top'
        self.add(self.if_space)
        self.conditional = blocks.BooleanEntryBlock(0,0)
        self.conditional.layout_hint = 'top'
        self.conditional_space = canvas.Space(0, 0, 6, 6)
        self.conditional_space.layout_hint = 'top'
        self.add(self.conditional)
        self.add(self.conditional_space)
        self.then_text = canvas.Text(0, 0, text[1], (0xFF,0xFF,0xFF,0xFF))
        self.then_text.layout_hint = 'top'
        self.add(self.then_text)
        self.else_text = canvas.Text(0, 0, text_middle[0], (0xFF,0xFF,0xFF,0xFF))
        self.else_text.layout_hint = 'middle'
        self.add(self.else_text)
        self.end_text=canvas.Text(0, 0, text_bottom[0], (0xFF,0xFF,0xFF,0xFF))
        self.end_text.layout_hint = 'bottom'
        self.add(self.end_text)
        self.render()

    def to_xml(self, name=None, position=False):
        element = blocks.XMLBlock.to_xml(self, name, position)
        element.append(self.conditional.to_xml('conditional'))
        return element

    def from_xml(self, element, canvas_dest):
        blocks.XMLBlock.from_xml(self, element, canvas_dest)
        self.conditional.from_xml(element.find('conditional'), canvas_dest)

    def basic_code(self, inits, setups):
        if self.conditional.adopt_point.attached_to:
            condition = self.conditional.adopt_point.attached_to.basic_code(inits, setups)
        else:
            return ''
        basic_code = 'if %s then\n' % condition
        if self.adopt_point_one.attached_to or self.adopt_point_two.attached_to:
            if self.adopt_point_one.attached_to:
                blocks = self.canvas.find_all_attached(self.adopt_point_one.attached_to)
                for block in blocks:
                    basic_code += block.basic_code(inits, setups)
                basic_code += 'else\n'
            else:
                basic_code += 'pass\nelse\n'
            if self.adopt_point_two.attached_to:
                blocks = self.canvas.find_all_attached(self.adopt_point_two.attached_to)
                for block in blocks:
                    basic_code += block.basic_code(inits, setups)
                basic_code += 'end if\n'
            else:
                basic_code += 'pass\nend if\n'

            return basic_code
        else:
            return ''


class DoUntilBlock(blocks.ControlBlock, blocks.XMLBlock):
    def __init__(self, x, y):
        super(DoUntilBlock, self).__init__(x, y)
        text = _('do until|').split('|')
        text_bottom = _('loop').split('|')
        self.loop_text = canvas.Text(0, 0, text[0], (0xFF,0xFF,0xFF,0xFF))
        self.loop_text.layout_hint = 'top'
        self.add(self.loop_text)
        self.loop_space = canvas.Space(0, 0, 6, 6)
        self.loop_space.layout_hint = 'top'
        self.add(self.loop_space)
        self.conditional = blocks.BooleanEntryBlock(0,0)
        self.conditional.layout_hint = 'top'
        self.add(self.conditional)
        self.next_text=canvas.Text(0, 0, text_bottom[0], (0xFF,0xFF,0xFF,0xFF))
        self.next_text.layout_hint = 'bottom'
        self.add(self.next_text)
        self.render()

    def to_xml(self, name=None, position=False):
        element = blocks.XMLBlock.to_xml(self, name, position)
        element.append(self.conditional.to_xml('conditional'))
        return element

    def from_xml(self, element, canvas_dest):
        blocks.XMLBlock.from_xml(self, element, canvas_dest)
        self.conditional.from_xml(element.find('conditional'), canvas_dest)

    def basic_code(self, inits, setups):
        if self.conditional.adopt_point.attached_to:
            condition = self.conditional.adopt_point.attached_to.basic_code(inits, setups)
        else:
            condition = '0'
        basic_code = 'do until %s\n' % condition
        if self.adopt_point.attached_to:
            blocks = self.canvas.find_all_attached(self.adopt_point.attached_to)
            for block in blocks:
                basic_code += block.basic_code(inits, setups)
            basic_code += 'loop\n'
            return basic_code
        else:
            return ''


def LoopUntilBlock(x, y):
    """For compatibility with 0.13.9 and earlier"""
    block = DoUntilBlock(x, y)
    return block


class DoTimesBlock(blocks.ControlBlock, blocks.XMLBlock):
    def __init__(self, x, y):
        super(DoTimesBlock, self).__init__(x, y)
        text = _('do|times').split('|')
        text_bottom = _('loop').split('|')
        self.loop_text = canvas.Text(0, 0, text[0], (0xFF,0xFF,0xFF,0xFF))
        self.loop_text.layout_hint = 'top'
        self.add(self.loop_text)
        self.loop_space = canvas.Space(0, 0, 6, 6)
        self.loop_space.layout_hint = 'top'
        self.add(self.loop_space)
        self.value = blocks.ValueEntryBlock(0, 0, 10)
        self.value.layout_hint = 'top'
        self.value_space = canvas.Space(0, 0, 6, 6)
        self.value_space.layout_hint = 'top'
        self.add(self.value)
        self.add(self.value_space)
        self.times_text = canvas.Text(0, 0, text[1], (0xFF,0xFF,0xFF,0xFF))
        self.times_text.layout_hint = 'top'
        self.add(self.times_text)
        self.next_text=canvas.Text(0, 0, text_bottom[0], (0xFF,0xFF,0xFF,0xFF))
        self.next_text.layout_hint = 'bottom'
        self.add(self.next_text)
        self.render()

    def to_xml(self, name=None, position=False):
        element = blocks.XMLBlock.to_xml(self, name, position)
        element.append(self.value.to_xml('count'))
        return element

    def from_xml(self, element, canvas_dest):
        blocks.XMLBlock.from_xml(self, element, canvas_dest)
        self.value.from_xml(element.find('count'), canvas_dest)

    def basic_code(self, inits, setups):
        basic_code = 'do %s times\n' % self.value.basic_code(inits, setups)
        if self.adopt_point.attached_to:
            blocks = self.canvas.find_all_attached(self.adopt_point.attached_to)
            for block in blocks:
                basic_code += block.basic_code(inits, setups)
            basic_code += 'loop\n'
            return basic_code
        else:
            return ''


class DoForeverBlock(blocks.ControlForeverBlock, blocks.XMLBlock):
    def __init__(self, x, y):
        super(DoForeverBlock, self).__init__(x, y)
        text = _('do forever').split('|')
        text_bottom = _('loop').split('|')
        self.loop_text = canvas.Text(0, 0, text[0], (0xFF,0xFF,0xFF,0xFF))
        self.loop_text.layout_hint = 'top'
        self.add(self.loop_text)
        self.next_text=canvas.Text(0, 0, text_bottom[0], (0xFF,0xFF,0xFF,0xFF))
        self.next_text.layout_hint = 'bottom'
        self.add(self.next_text)
        self.render()

    def basic_code(self, inits, setups):
        basic_code = 'do forever\n'
        blocks = self.canvas.find_all_attached(self.adopt_point.attached_to)
        for block in blocks:
            basic_code += block.basic_code(inits, setups)
        basic_code += 'loop\n'
        return basic_code


def LoopForeverBlock(x, y):
    """For compatibility with 0.13.9 and earlier"""
    block = DoForeverBlock(x, y)
    return block

class AssignmentBlock(blocks.DataStatementBlock, blocks.XMLBlock):
    def __init__(self, x, y):
        super(AssignmentBlock, self).__init__(x, y)
        text = _('let|=|').split('|')
        self.add(canvas.Text(0, 0, text[0], (255,255,255,255)))
        self.add(canvas.Space(0, 0 , 6, 5))
        self.variable_entry = blocks.VariableEntryBlock(0, 0)
        self.add(self.variable_entry)
        self.add(canvas.Space(0, 0 , 6, 5))
        self.add(canvas.Text(0, 0, text[1], (0xFF,0xFF,0xFF,0xFF)))
        self.add(canvas.Space(0, 0 , 6, 5))
        self.expression_entry = blocks.ValueEntryBlock(0, 0, 0)
        self.add(self.expression_entry)
        self.render()

    def to_xml(self, name=None, position=False):
        element = blocks.XMLBlock.to_xml(self, name, position)
        element.append(self.variable_entry.to_xml('larg'))
        element.append(self.expression_entry.to_xml('rarg'))
        return element

    def from_xml(self, element, canvas_dest):
        blocks.XMLBlock.from_xml(self, element, canvas_dest)
        self.variable_entry.from_xml(element.find('larg'), canvas_dest)
        self.expression_entry.from_xml(element.find('rarg'), canvas_dest)

    def basic_code(self, inits, setups):
        variable = self.variable_entry.basic_code(inits, setups)
        if variable:
            return '%s = %s\n' % (variable, self.expression_entry.basic_code(inits, setups))
        else:
            return ''


class IncreaseBlock(blocks.DataStatementBlock, blocks.XMLBlock):
    def __init__(self, x, y):
        super(IncreaseBlock, self).__init__(x, y)
        text = _('increase|by|').split('|')
        self.add(canvas.Text(0, 0, text[0], (255,255,255,255)))
        self.add(canvas.Space(0, 0 , 6, 5))
        self.variable_entry = blocks.VariableEntryBlock(0, 0)
        self.add(self.variable_entry)
        self.add(canvas.Space(0, 0 , 6, 5))
        self.add(canvas.Text(0, 0, text[1], (0xFF,0xFF,0xFF,0xFF)))
        self.add(canvas.Space(0, 0 , 6, 5))
        self.expression_entry = blocks.ValueEntryBlock(0, 0, 1)
        self.add(self.expression_entry)
        self.render()

    def to_xml(self, name=None, position=False):
        element = blocks.XMLBlock.to_xml(self, name, position)
        element.append(self.variable_entry.to_xml('larg'))
        element.append(self.expression_entry.to_xml('rarg'))
        return element

    def from_xml(self, element, canvas_dest):
        blocks.XMLBlock.from_xml(self, element, canvas_dest)
        self.variable_entry.from_xml(element.find('larg'), canvas_dest)
        self.expression_entry.from_xml(element.find('rarg'), canvas_dest)

    def basic_code(self, inits, setups):
        variable = self.variable_entry.basic_code(inits, setups)
        if variable:
            return '%s = %s + %s\n' % (variable, variable, self.expression_entry.basic_code(inits, setups))
        else:
            return ''


class DecreaseBlock(blocks.DataStatementBlock, blocks.XMLBlock):
    def __init__(self, x, y):
        super(DecreaseBlock, self).__init__(x, y)
        text = _('decrease|by|').split('|')
        self.add(canvas.Text(0, 0, text[0], (255,255,255,255)))
        self.add(canvas.Space(0, 0 , 6, 5))
        self.variable_entry = blocks.VariableEntryBlock(0, 0)
        self.add(self.variable_entry)
        self.add(canvas.Space(0, 0 , 6, 5))
        self.add(canvas.Text(0, 0, text[1], (0xFF,0xFF,0xFF,0xFF)))
        self.add(canvas.Space(0, 0 , 6, 5))
        self.expression_entry = blocks.ValueEntryBlock(0, 0, 1)
        self.add(self.expression_entry)
        self.render()

    def to_xml(self, name=None, position=False):
        element = blocks.XMLBlock.to_xml(self, name, position)
        element.append(self.variable_entry.to_xml('larg'))
        element.append(self.expression_entry.to_xml('rarg'))
        return element

    def from_xml(self, element, canvas_dest):
        blocks.XMLBlock.from_xml(self, element, canvas_dest)
        self.variable_entry.from_xml(element.find('larg'), canvas_dest)
        self.expression_entry.from_xml(element.find('rarg'), canvas_dest)

    def basic_code(self, inits, setups):
        variable = self.variable_entry.basic_code(inits, setups)
        if variable:
            return '%s = %s - %s\n' % (variable, variable, self.expression_entry.basic_code(inits, setups))
        else:
            return ''


class VariableBlock(blocks.ValueBlock, blocks.XMLBlock):
    # instances = WeakSet()
    def __init__(self, x, y):
        super(VariableBlock, self).__init__(x, y)
        # VariableBlock.instances.add(self)
        self.type_string = 'Variable'
        self.name_text = canvas.Text(0, 0, '', (0xFF,0xFF,0xFF,0xFF))
        self.add(self.name_text)
        self._var_name = None
        pub.subscribe(self.rename_listener, 'crumble.variable.rename')

    @property
    def var_name(self):
        return self._var_name

    @var_name.setter
    def var_name(self, var_name):
        self._var_name = var_name
        self.name_text.text = var_name
        self.render()

    def rename_listener(self, old_name, var_name):
        if self.var_name == old_name:
            self.var_name = var_name

    def rename(self):
        blocks.VarNameEntryCtrl(self.canvas, self.canvas.calc_canvas_position(self) + wx.Point2D(0, self.height + 2), self.var_name)

    def spawn(self):
        new_block = super(VariableBlock, self).spawn()
        new_block.var_name = self.var_name
        return new_block

    def to_xml(self, name=None, position=False):
        element = blocks.XMLBlock.to_xml(self, name, position)
        element.text = str(self.var_name.replace(' ', '_'))
        return element

    def from_xml(self, element, canvas_dest):
        if element is not None:
            blocks.XMLBlock.from_xml(self, element, canvas_dest)
            self.var_name = element.text.replace('_', ' ')

    def basic_code(self, inits, setups):
        spaceless_name = self.var_name.replace(' ', '_')
        inits['int %s\n' % spaceless_name] = 1
        return spaceless_name


class VariableMonitorBlock(blocks.MonitorBlock, blocks.XMLBlock):
    def __init__(self, x, y):
        super(VariableMonitorBlock, self).__init__(x, y)
        self.type_string = 'VariableMonitor'
        self.value_text = canvas.Text(0, 0, '?', (0x00,0x00,0x00,0xFF))
        self.add(self.value_text)
        self._value = None
        self._var_name = None
        self.render()
        pub.subscribe(self.monitor_stopped_listener, 'crumble.monitor.stopped')
        pub.subscribe(self.rename_listener, 'crumble.variable.rename')
        pub.subscribe(self.values_listener, 'crumble.variable.values')

    def values_listener(self, values):
        if self.var_name in values.keys():
            self.value = values[self.var_name]

    def rename_listener(self, old_name, var_name):
        if self.var_name == old_name:
            self.var_name = var_name

    def monitor_stopped_listener(self):
        self._value = None
        self.value_text.text = '?'
        self.render()

    @property
    def var_name(self):
        return self._var_name

    @var_name.setter
    def var_name(self, var_name):
        self._var_name = var_name

    @property
    def value(self):
        return self._value

    @value.setter
    def value(self, value):
        if value!=self._value:
            self._value = value
            self.value_text.text = str(value)
            self.render()


class LessThanBlock(blocks.BooleanBlock, blocks.XMLBlock):
    def __init__(self, x, y):
        super(LessThanBlock, self).__init__(x, y)
        self.left_entry = blocks.ValueEntryBlock(0, 0, 0)
        self.add(self.left_entry)
        self.add(canvas.Space(0, 0 , 6, 5))
        self.add(canvas.Text(0, 0, '<', (255,255,255,255)))
        self.add(canvas.Space(0, 0 , 6, 5))
        self.right_entry = blocks.ValueEntryBlock(0, 0, 0)
        self.add(self.right_entry)
        self.render()

    def to_xml(self, name=None, position=False):
        element = blocks.XMLBlock.to_xml(self, name, position)
        element.append(self.left_entry.to_xml('larg'))
        element.append(self.right_entry.to_xml('rarg'))
        return element

    def from_xml(self, element, canvas_dest):
        blocks.XMLBlock.from_xml(self, element, canvas_dest)
        self.left_entry.from_xml(element.find('larg'), canvas_dest)
        self.right_entry.from_xml(element.find('rarg'), canvas_dest)

    def basic_code(self, inits, setups):
        return '(%s < %s)' % (self.left_entry.basic_code(inits, setups), self.right_entry.basic_code(inits, setups))


class GreaterThanBlock(blocks.BooleanBlock, blocks.XMLBlock):
    def __init__(self, x, y):
        super(GreaterThanBlock, self).__init__(x, y)
        self.left_entry = blocks.ValueEntryBlock(0, 0, 0)
        self.add(self.left_entry)
        self.add(canvas.Space(0, 0 , 6, 5))
        self.add(canvas.Text(0, 0, '>', (255,255,255,255)))
        self.add(canvas.Space(0, 0 , 6, 5))
        self.right_entry = blocks.ValueEntryBlock(0, 0, 0)
        self.add(self.right_entry)
        self.render()

    def to_xml(self, name=None, position=False):
        element = blocks.XMLBlock.to_xml(self, name, position)
        element.append(self.left_entry.to_xml('larg'))
        element.append(self.right_entry.to_xml('rarg'))
        return element

    def from_xml(self, element, canvas_dest):
        blocks.XMLBlock.from_xml(self, element, canvas_dest)
        self.left_entry.from_xml(element.find('larg'), canvas_dest)
        self.right_entry.from_xml(element.find('rarg'), canvas_dest)

    def basic_code(self, inits, setups):
        return '(%s > %s)' % (self.left_entry.basic_code(inits, setups), self.right_entry.basic_code(inits, setups))


class EqualityBlock(blocks.BooleanBlock, blocks.XMLBlock):
    def __init__(self, x, y):
        super(EqualityBlock, self).__init__(x, y)
        self.left_entry = blocks.ValueEntryBlock(0, 0, 0)
        self.add(self.left_entry)
        self.add(canvas.Space(0, 0 , 6, 5))
        self.add(canvas.Text(0, 0, '=', (255,255,255,255)))
        self.add(canvas.Space(0, 0 , 6, 5))
        self.right_entry = blocks.ValueEntryBlock(0, 0, 0)
        self.add(self.right_entry)
        self.render()

    def to_xml(self, name=None, position=False):
        element = blocks.XMLBlock.to_xml(self, name, position)
        element.append(self.left_entry.to_xml('larg'))
        element.append(self.right_entry.to_xml('rarg'))
        return element

    def from_xml(self, element, canvas_dest):
        blocks.XMLBlock.from_xml(self, element, canvas_dest)
        self.left_entry.from_xml(element.find('larg'), canvas_dest)
        self.right_entry.from_xml(element.find('rarg'), canvas_dest)

    def basic_code(self, inits, setups):
        return '(%s == %s)' % (self.left_entry.basic_code(inits, setups), self.right_entry.basic_code(inits, setups))


class NotEqualBlock(blocks.BooleanBlock, blocks.XMLBlock):
    def __init__(self, x, y):
        super(NotEqualBlock, self).__init__(x, y)
        self.left_entry = blocks.ValueEntryBlock(0, 0, 0)
        self.add(self.left_entry)
        self.add(canvas.Space(0, 0 , 6, 5))
        self.add(canvas.Text(0, 0, u'\u2260', (255,255,255,255)))
        self.add(canvas.Space(0, 0 , 6, 5))
        self.right_entry = blocks.ValueEntryBlock(0, 0, 0)
        self.add(self.right_entry)
        self.render()

    def to_xml(self, name=None, position=False):
        element = blocks.XMLBlock.to_xml(self, name, position)
        element.append(self.left_entry.to_xml('larg'))
        element.append(self.right_entry.to_xml('rarg'))
        return element

    def from_xml(self, element, canvas_dest):
        blocks.XMLBlock.from_xml(self, element, canvas_dest)
        self.left_entry.from_xml(element.find('larg'), canvas_dest)
        self.right_entry.from_xml(element.find('rarg'), canvas_dest)

    def basic_code(self, inits, setups):
        return '(not (%s == %s))' % (self.left_entry.basic_code(inits, setups), self.right_entry.basic_code(inits, setups))


class AndBlock(blocks.BooleanBlock, blocks.XMLBlock):
    def __init__(self, x, y):
        super(AndBlock, self).__init__(x, y)
        self.left_entry = blocks.BooleanEntryBlock(0, 0)
        text = _('|and|').split('|')
        self.add(self.left_entry)
        self.add(canvas.Space(0, 0 , 6, 5))
        self.add(canvas.Text(0, 0, text[1], (255,255,255,255)))
        self.add(canvas.Space(0, 0 , 6, 5))
        self.right_entry = blocks.BooleanEntryBlock(0, 0)
        self.add(self.right_entry)
        self.render()

    def to_xml(self, name=None, position=False):
        element = blocks.XMLBlock.to_xml(self, name, position)
        element.append(self.left_entry.to_xml('larg'))
        element.append(self.right_entry.to_xml('rarg'))
        return element

    def from_xml(self, element, canvas_dest):
        blocks.XMLBlock.from_xml(self, element, canvas_dest)
        self.left_entry.from_xml(element.find('larg'), canvas_dest)
        self.right_entry.from_xml(element.find('rarg'), canvas_dest)

    def basic_code(self, inits, setups):
        return '(%s and %s)' % (self.left_entry.basic_code(inits, setups), self.right_entry.basic_code(inits, setups))


class OrBlock(blocks.BooleanBlock, blocks.XMLBlock):
    def __init__(self, x, y):
        super(OrBlock, self).__init__(x, y)
        text = _('|or|').split('|')
        self.left_entry = blocks.BooleanEntryBlock(0, 0)
        self.add(self.left_entry)
        self.add(canvas.Space(0, 0 , 6, 5))
        self.add(canvas.Text(0, 0, text[1], (255,255,255,255)))
        self.add(canvas.Space(0, 0 , 6, 5))
        self.right_entry = blocks.BooleanEntryBlock(0, 0)
        self.add(self.right_entry)
        self.render()

    def to_xml(self, name=None, position=False):
        element = blocks.XMLBlock.to_xml(self, name, position)
        element.append(self.left_entry.to_xml('larg'))
        element.append(self.right_entry.to_xml('rarg'))
        return element

    def from_xml(self, element, canvas_dest):
        blocks.XMLBlock.from_xml(self, element, canvas_dest)
        self.left_entry.from_xml(element.find('larg'), canvas_dest)
        self.right_entry.from_xml(element.find('rarg'), canvas_dest)

    def basic_code(self, inits, setups):
        return '(%s or %s)' % (self.left_entry.basic_code(inits, setups), self.right_entry.basic_code(inits, setups))


class NotBlock(blocks.BooleanBlock, blocks.XMLBlock):
    def __init__(self, x, y):
        super(NotBlock, self).__init__(x, y)
        text = _('|not|').split('|')
        self.add(canvas.Text(0, 0, text[1], (255,255,255,255)))
        self.add(canvas.Space(0, 0 , 6, 5))
        self.right_entry = blocks.BooleanEntryBlock(0, 0)
        self.add(self.right_entry)
        self.render()

    def to_xml(self, name=None, position=False):
        element = blocks.XMLBlock.to_xml(self, name, position)
        element.append(self.right_entry.to_xml('rarg'))
        return element

    def from_xml(self, element, canvas_dest):
        blocks.XMLBlock.from_xml(self, element, canvas_dest)
        self.right_entry.from_xml(element.find('rarg'), canvas_dest)

    def basic_code(self, inits, setups):
        return '(not %s)' % self.right_entry.basic_code(inits, setups)


class AddBlock(blocks.OperatorBlock, blocks.XMLBlock):
    def __init__(self, x, y):
        super(AddBlock, self).__init__(x, y)
        self.left_entry = blocks.ValueEntryBlock(0, 0, 0)
        self.add(self.left_entry)
        self.add(canvas.Space(0, 0 , 6, 5))
        self.add(canvas.Text(0, 0, '+', (255,255,255,255)))
        self.add(canvas.Space(0, 0 , 6, 5))
        self.right_entry = blocks.ValueEntryBlock(0, 0, 0)
        self.add(self.right_entry)
        self.render()

    def to_xml(self, name=None, position=False):
        element = blocks.XMLBlock.to_xml(self, name, position)
        element.append(self.left_entry.to_xml('larg'))
        element.append(self.right_entry.to_xml('rarg'))
        return element

    def from_xml(self, element, canvas_dest):
        blocks.XMLBlock.from_xml(self, element, canvas_dest)
        self.left_entry.from_xml(element.find('larg'), canvas_dest)
        self.right_entry.from_xml(element.find('rarg'), canvas_dest)

    def basic_code(self, inits, setups):
        return '(%s + %s)' % (self.left_entry.basic_code(inits, setups), self.right_entry.basic_code(inits, setups))


class SubBlock(blocks.OperatorBlock, blocks.XMLBlock):
    def __init__(self, x, y):
        super(SubBlock, self).__init__(x, y)
        self.left_entry = blocks.ValueEntryBlock(0, 0, 0)
        self.add(self.left_entry)
        self.add(canvas.Space(0, 0 , 6, 5))
        self.add(canvas.Text(0, 0, '-', (255,255,255,255)))
        self.add(canvas.Space(0, 0 , 6, 5))
        self.right_entry = blocks.ValueEntryBlock(0, 0, 0)
        self.add(self.right_entry)
        self.render()

    def to_xml(self, name=None, position=False):
        element = blocks.XMLBlock.to_xml(self, name, position)
        element.append(self.left_entry.to_xml('larg'))
        element.append(self.right_entry.to_xml('rarg'))
        return element

    def from_xml(self, element, canvas_dest):
        blocks.XMLBlock.from_xml(self, element, canvas_dest)
        self.left_entry.from_xml(element.find('larg'), canvas_dest)
        self.right_entry.from_xml(element.find('rarg'), canvas_dest)

    def basic_code(self, inits, setups):
        return '(%s - %s)' % (self.left_entry.basic_code(inits, setups), self.right_entry.basic_code(inits, setups))


class MultiplyBlock(blocks.OperatorBlock, blocks.XMLBlock):
    def __init__(self, x, y):
        super(MultiplyBlock, self).__init__(x, y)
        self.left_entry = blocks.ValueEntryBlock(0, 0, 0)
        self.add(self.left_entry)
        self.add(canvas.Space(0, 0 , 6, 5))
        self.add(canvas.Text(0, 0, u'\u00d7', (255,255,255,255)))
        self.add(canvas.Space(0, 0 , 6, 5))
        self.right_entry = blocks.ValueEntryBlock(0, 0, 0)
        self.add(self.right_entry)
        self.render()

    def to_xml(self, name=None, position=False):
        element = blocks.XMLBlock.to_xml(self, name, position)
        element.append(self.left_entry.to_xml('larg'))
        element.append(self.right_entry.to_xml('rarg'))
        return element

    def from_xml(self, element, canvas_dest):
        blocks.XMLBlock.from_xml(self, element, canvas_dest)
        self.left_entry.from_xml(element.find('larg'), canvas_dest)
        self.right_entry.from_xml(element.find('rarg'), canvas_dest)


    def basic_code(self, inits, setups):
        return '(%s * %s)' % (self.left_entry.basic_code(inits, setups), self.right_entry.basic_code(inits, setups))


class DivideBlock(blocks.OperatorBlock, blocks.XMLBlock):
    def __init__(self, x, y):
        super(DivideBlock, self).__init__(x, y)
        self.left_entry = blocks.ValueEntryBlock(0, 0, 0)
        self.add(self.left_entry)
        self.add(canvas.Space(0, 0 , 6, 5))
        self.add(canvas.Text(0, 0, u'\xf7', (255,255,255,255)))
        self.add(canvas.Space(0, 0 , 6, 5))
        self.right_entry = blocks.ValueEntryBlock(0, 0, 0)
        self.add(self.right_entry)
        self.render()

    def to_xml(self, name=None, position=False):
        element = blocks.XMLBlock.to_xml(self, name, position)
        element.append(self.left_entry.to_xml('larg'))
        element.append(self.right_entry.to_xml('rarg'))
        return element

    def from_xml(self, element, canvas_dest):
        blocks.XMLBlock.from_xml(self, element, canvas_dest)
        self.left_entry.from_xml(element.find('larg'), canvas_dest)
        self.right_entry.from_xml(element.find('rarg'), canvas_dest)

    def basic_code(self, inits, setups):
        return '(%s / %s)' % (self.left_entry.basic_code(inits, setups), self.right_entry.basic_code(inits, setups))


class RandomBlock(blocks.OperatorBlock, blocks.XMLBlock):
    def __init__(self, x, y):
        super(RandomBlock, self).__init__(x, y)
        text = _('random|to|').split('|')
        self.add(canvas.Space(0, 0, 6, 5))
        self.add(canvas.Text(0, 0, text[0], (255,255,255,255)))
        self.add(canvas.Space(0, 0, 6, 5))
        self.min_entry = blocks.ValueEntryBlock(0, 0, 0)
        self.add(self.min_entry)
        self.add(canvas.Space(0, 0, 6, 5))
        self.add(canvas.Text(0, 0, text[1], (255,255,255,255)))
        self.add(canvas.Space(0, 0, 6, 5))
        self.max_entry = blocks.ValueEntryBlock(0, 0, 10)
        self.add(self.max_entry)
        self.add(canvas.Space(0, 0, 6, 5))
        self.add(canvas.Text(0, 0, text[2], (255,255,255,255)))
        if text[2]:
            self.add(canvas.Space(0, 0, 6, 5))
        self.render()

    def to_xml(self, name=None, position=False):
        element = blocks.XMLBlock.to_xml(self, name, position)
        element.append(self.min_entry.to_xml('min'))
        element.append(self.max_entry.to_xml('max'))
        return element

    def from_xml(self, element, canvas_dest):
        blocks.XMLBlock.from_xml(self, element, canvas_dest)
        self.min_entry.from_xml(element.find('min'), canvas_dest)
        self.max_entry.from_xml(element.find('max'), canvas_dest)

    def basic_code(self, inits, setups):
        return 'Random(%s, %s)' % (self.min_entry.basic_code(inits, setups), self.max_entry.basic_code(inits, setups))